#Date:    20 Nov 1995 20:33:38 +0000
#From:    Robin Friedrich <friedric@rose.rsoc.rockwell.com>
#To:      python-list@cwi.nl
#Subject: HTMLsupport.py
#
#Just thought I might post this thing up again. Someone asked for
#simple examples of python code and I find this enhanced HTMLsupport
#module useful for generating web pages.  It includes a class for 
#creating Netscape tables from a list of lists. Real handy.
#
#Bon appetite!
#
#
"""HTMLsupport is a collection of simple utility functions useful for
generating HTML docs from Python scripts. For brevity it's intended
to be used as 'from HTMLsupport import *' (don't shoot me).
Also included is a HTMLtable class designed to easily generate tables
using Netscape 1.1 syntax. It simply takes a list of lists.
This module is derived from a like named module whose origin escapes
me. It's a superset of the original so a replacement should be safe.
I've also thrown in some useful color constants at the end.
This is by no means polished but I find it useful and maybe others
will too.  I generate gobs of tables with this thing.

                     Robin Friedrich
                     friedric@rose.rsoc.rockwell.com
"""
import string

def IMG(fd,fn):
    fd.write('<IMG ALIGN=middle SRC=\"'+fn+'\">')

def DOCSTART(fd, contentType = '', title = '', BG=None, TC=None):
    if contentType:
      CONTENTTYPE(fd, contentType)
    fd.write('<HTML>')
    if title:
      fd.write('<HEAD>\n')
      TITLE(fd, title)
      fd.write('</HEAD>')
    BODYSTART(fd, BG, TC)

def DOCEND(fd):
    fd.write('</HTML>\n')

def CONTENTTYPE(fd, content):
    fd.write('Content-type: ' + content + '\n\n')

def TITLE(fd,s):
    fd.write('<TITLE>'+s+'</TITLE>\n')

def BODYSTART(fd, BG=None, TC=None):
    bgsetting, tcsetting = '', ''
    if BG:
	if '.' in BG: #he looks to be passing a file name; set wallpaper
	    bgsetting = ' BACKGROUND="'+BG+'" '
	else: #try setting to flat color background
	    bgsetting = ' BGCOLOR="'+BG+'" '
    if TC:
	tcsetting = ' TEXT="'+TC+'"'
    fd.write('<BODY '+ bgsetting + tcsetting + '>\n')


def BODYEND(fd):
    fd.write('</BODY>\n')

def H1(fd,s):
    fd.write('<H1>'+s+'</H1>\n')

def H2(fd,s):
    fd.write('<H2>'+s+'</H2>\n')

def H3(fd,s):
    fd.write('<H3>'+s+'</H3>\n')

def H4(fd,s):
    fd.write('<H4>'+s+'</H3>\n')

def DL(fd):
    fd.write('<DL><DL>\n')

def DESC(fd,s,text):
    fd.write('<DT>'+s+'\n<DD>'+text)

def DLE(fd):
    fd.write('</DL></DL>\n')

def LIST(fd, list):
    fd.write('<UL>\n')
    for item in list:
	fd.write('<LI>'+item)
    fd.write('</UL>\n')

def LISTN(fd, list):
    fd.write('<OL>\n')
    for item in list:
	fd.write('<LI>'+item)
    fd.write('</OL>\n')

def PRE(fd,s):
    fd.write('<PRE>'+s+'</PRE>\n')

def AN(fd,tag,s):
    fd.write('<A NAME="'+tag+'"> '+s+'</A>\n')

def AH(fd,s1,s2):
    fd.write('<A HREF="'+s1+'"> '+s2+'</A>\n')

def PAR(fd):
    P(fd)

def P(fd):
    fd.write('<P>\n')

def BR(fd):
    fd.write('<BR>\n')

def HR(fd, thickness = 1):
    fd.write('<HR size='+`thickness`+'>\n')

def PHR(fd, thickness = 1):
    fd.write('<P>\n<HR size='+`thickness`+'>\n')

#Character Highlighting
#
def EM(fd, s):
    fd.write('<EM>'+s+'</EM>')

def STRONG(fd, s):
    fd.write('<STRONG>'+s+'</STRONG>')

def CODE(fd, s):
    fd.write('<CODE>'+s+'</CODE>')

class HTMLtable:
    """Usage:
    Instantiate with a string argument for the table's name (caption).
    Set object.heading to a list of strings representing the column headings.
    Set object.body to a list of lists representing rows.(must be strings)
    Then just give an open file object for it to write to...   object.write(f)
    This only generates the actual table, the user is reponsible for generating
    the html body and closures.
    """
    def __init__(self, title = ''):
	self.title = title
	self.hypername = '<A NAME="'+self.title+'"></A>'
	self.props = "<TABLE border=1 cellpadding=4 cellspacing=1>"
	self.heading = ['&nbsp']*5
	self.body = [['&nbsp']*5]
	self.cell_alignment = 'right'

    def __repr__(self):
	s = stringer()
	self.write(s)
	return s.read()

    def write(self, f):
	f.write(self.hypername+self.props+'\n')
	f.write('<CAPTION align=top> <B>'+self.title+'</B> </CAPTION>\n')
	f.write('<TH>'+string.joinfields(self.heading,'<TH>')+'<TR>\n')
	for bodyline in self.body:
	    f.write('<TD VALIGN="top"> '+\
		    string.joinfields(bodyline,
			  '<TD VALIGN="top" align="'+self.cell_alignment+'">')+
\
		    '<TR>\n')
	f.write('</TABLE><P>\n')

class stringer:
    """An instance of this class can act as an open file handle.
    The bytes sent to the write method are stored in a string buffer.
    The read method can be used to return the buffer as a string.
    Optionally, an integer can be given to the read method and that
    many characters from the front of the buffer are returned (and
    truncated off the front of the buffer).
    """
    def __init__(self):
	self.buffer = ''
    def write(self, string):
	self.buffer = self.buffer + string
    def read(self, chars= None):
	if chars == None:
	    return self.buffer
	else:
	    retstr = self.buffer[:chars]
	    self.buffer = self.buffer[chars:]
	    return retstr


#SELECTED COLORS USEFUL FOR NETSCAPE 1.1 HTML BACKGROUNDS/FOREGROUNDS.
# These are some hex triplet codes corresponding to colors which are displayed
# without dithering.  Numbered colors get darker as their numbers increase.
# There are 256 such colors in Netscape's colorcube but only distinct colors
# are set here.
GREY1,     GREY2,    GREY3,     GREY4,   GREY5,    GREY6 = \
'#EEEEEE','#DDDDDD','#CCCCCC','#BBBBBB','#AAAAAA','#999999'
RED1,      RED2,     RED3,     RED4,     RED5,     RED6 = \
'#FF0000','#EE0000','#DD0000','#CC0000','#BB0000','#990000'
BLUE1,     BLUE2,    BLUE3,    BLUE4,    BLUE5,    BLUE6 = \
'#0000FF','#0000EE','#0000DD','#0000CC','#0000BB','#000099'
GREEN1,    GREEN2,   GREEN3,   GREEN4,   GREEN5,   GREEN6 = \
'#00EE00','#00DD00','#00CC00','#00BB00','#00AA00','#009900'
RED = RED2
GREEN = GREEN2
BLUE = BLUE1
GREY = GRAY = GREY3
WHITE = '#FFFFFF'
BLACK = '#000000'
ORANGE = '#FFCC33'
SKYBLUE = '#99CCFF'
BROWN = '#CC9966'
SEAGREEN = '#0099CC'
MOROON = '#CC3366'
CADETBLUE = '#9999CC'
AQUA = '#33CCCC'
YELLOW = '#FFFF66'
PURPLE = '#CC66FF'
PEACH = '#FFCC99'
LIME = '#CCFF66'
OLIVE = '#999966'
TAN = '#FFCCCC'
PINK = '#FF99CC'



