################################################################################
#
# File:         bibdocxform.py
# RCS:          $Header: /u/testbed/CVSROOT/dldev/src/interbib2007/Scripts/bibdocxform.py,v 1.1 2007/07/24 20:24:57 paepcke Exp $
# Description:  Reading documents and creating new ones with bibliographies added
# Author:       Andreas Paepcke, Stanford University
# Created:      Sun Oct 15 07:36:53 1995
# Modified:     Mon Jul 23 17:39:39 2007 (Andreas Paepcke) paepcke@SNUGGLES
# Language:     Python
# Package:      N/A
# Status:       Experimental (Do Not Distribute)
#
# (c) Copyright 1995, Stanford University, all rights reserved.
#
################################################################################
#
# Revisions:
#
# Fri Jul 20 11:37:07 2007 (Andreas Paepcke) paepcke@Shark.Stanford.EDU
#  Migrated from deprecated regex to re.
# Thu Sep 19 13:15:47 1996 (Andreas Paepcke) paepcke@Walrus
#  Fixed to work with Macintosh files
# Tue Sep 17 13:59:39 1996 (Andreas Paepcke) paepcke@Walrus
#  Fixed error caused when transformDocument got a bibl insertion string
#  that was a special char for regular expressions.
# Wed Sep 11 17:00:53 1996 (Andreas Paepcke) paepcke@Walrus
#  Added HTML support
# Sun May  5 10:53:49 1996 (Tom Schirmer) schirmer@cs.Stanford.EDU
#  added RTF support
# Tue Feb 20 19:51:01 1996 (Andreas Paepcke) paepcke@Stingray
#  Added AuthorCiteAuthorAlphaOrder and AuthorCiteAppearanceOrder styles
# Sat Feb  3 22:18:38 1996 (Andreas Paepcke) paepcke@AndreasHomePC
#  Added option to include/exclude annotations in bibliographies.
# Fri Jan 26 13:43:30 1996 (Andreas Paepcke) paepcke@Stingray
#  Allowed multiple chars in citation opener
# Fri Jan 26 09:23:48 1996 (Andreas Paepcke) paepcke@Stingray
#  Added feedback output.
# Sun Jan 14 09:04:05 1996 (Andreas Paepcke) paepcke@AndreasHomePC
#  Added method setDocumentCitationDelimiters
# Sun Jan  7 11:11:03 1996 (Andreas Paepcke) paepcke@AndreasHomePC
#  When detecting key spec in multiple dbs, now identifies db by name
#  in warning.
# Thu Dec 28 09:58:55 1995 (Andreas Paepcke) paepcke@AndreasHomePC
#  Made transformDocument accept an FD for output.
# Tue Dec 19 18:42:00 1995 (Andreas Paepcke) paepcke@Phoebe.Stanford.EDU
#  Fixed bug that happened when citations went across a line break.
# Tue Dec 19 16:45:48 1995 (Andreas Paepcke) paepcke@Stingray
#  Added ability to specify an insertion point for the bibliography in MIF files.
################################################################################

# To do:

#   - In citation style: do the platform-dependent linking (do it in main
#     doc xform method: call compiler's linkOpen/linkClose methods around
#     the citeKey string.

# Public classes:

# Document transformers (These resolve citations and add reference lists):
#
#   - MIFDocTransformer (Takes and returns a MIF document)
#       transformDocument
#   - RTFDocTransformer (Takes and returns a RTF document)
#       transformDocument
#   - HTMLDocTransformer (Takes and returns an HTML document)
#       transformDocument
#
# Citation styles (Styles for the citations in the text and labels in ref list):
#
#   - NumericCiteAppearanceOrder
#   - AlphaCiteAppearanceOrder
#   - AuthorCiteAppearanceOrder
#   - AlphaCiteKeyAlphaOrder
#   - AlphaCiteAuthorAlphaOrder
#   - NumericCiteAuthorAlphaOrder
#   - AuthorCiteAuthorAlphaOrder
#
# (for Reference styles (i.e. appearance of references, see module bibrefgen)).

import string
import bibutils
import bibrefgen
import bibcompile
import bibconvert
import bibimport
import types
import tempfile
import os
import re

class DocTransformer:
  errors = []
  warnings = []
  verbosity = 1
  inFd = 0
  outFd = 0
  outStr = ''           # Used if we are outputting to a str, rather than a file
  refedEntries = []
  refList = []
  citationOpener = '['  # String opening a citation in raw doc (may be several chars long)
  citationCloser = ']'  # Closing *char* of a citation (raw doc). Could allow longer ones
                        # if needed
  citeKeyDelimiters = '' # chars that might delimit a citation key
  dbList = []           # List of citation dbs to search for bib entries
  eof    = 0
  cursor = 0            # used for buffering of file pieces
  buf    = ''           # used for buffering of file pieces
  
  # finalProductOutFd is where the final doc gets written to if caller
  # specifies it and if bibliography is inserted somewhere inside the
  # file. It should also work if the bibl is just to be inserted at the
  # end. But I don't have the patience right now to look for where to
  # splice that in.  The same goes for systemFileHeader: only for the
  # bibliography insertion case will I prepend the systemFileHeader. The
  # finalOutFd is used independently for one/two pass processing:

  def __init__(self, verbosity = 1):
    self.verbosity = verbosity
    self.refedEntries = []
    self.refList = []
#   self.citeKeyDelimiters = ', ;\n' + self.citationCloser
    self.citeKeyDelimiters = ', ;\n\r' + self.citationCloser
    self.outStr = ''
    self.eof = 0
    self.citationStartPointers = []
    self.finalOutFd = None
    self.finalProductOutFd = None
    self.systemFileHeader = ''
    self.tmpName = ''
    self.macFileTmpName = ''
    self.citationStyle = None
    self.twoPasses = 0
    self.bibliographyInsertPt = 0
    self.foundBibInsertPt = 0
    self.insertPtMatcher = 0
    self.tab = bibconvert.FrameTAB

  def handleWarning(self, warnStr):
    self.warnings.append(warnStr)
    if self.verbosity:
      print "Warning: " + warnStr

  def handleError(self, errStr, exception=0, excArg=""):
    self.errors.append(errStr)
    if self.verbosity:
      print 'Error: ' + self.errors[-1]
    if exception:
      raise exception, excArg

  def handleMissingRef(self, citeKey):
    self.handleWarning("Citation key '" +\
		       citeKey +\
		       "' not found in any of the databases")

  def openDocument(self, inFileName='', inDocStr=''):
    self.inFd = 0
    try:
      if inFileName:
        self.inFd = bibutils.StrOrFileFd(fileName=inFileName)
	self.checkForMacFile()
      else:
        self.inFd = bibutils.StrOrFileFd(str=inDocStr)
    except IOError, info:
      self.handleError("Input file could not be opened:", IOError, info)

  # Mac files have lines separated by '\015' (^M) only, while Unix is '\012' (^J).
  # If the file we are getting is a Mac file, we work in memory:

  def checkForMacFile(self):
    fileSize = self.inFd.len()
    # For Mac files, the following will be the entire thing:
    firstLn = self.inFd.readline()
    if len(firstLn) == fileSize and\
       string.find(firstLn, '\015'):
      unixFormatStr = re.sub.gsub('\015', '\012', firstLn)
      # Don't need the file anymore, because the whole contents is in the string:
      self.inFd.close()
      self.macFileTmpName = tempfile.mktemp() + '.mif'
      self.inFd = open(self.macFileTmpName, 'w')
      self.inFd.write(unixFormatStr)
      self.inFd.close()
      self.inFd = bibutils.StrOrFileFd(fileName=self.macFileTmpName)
    else:
      self.inFd.seek(0)

  def readBuf(self, size):
    res = ''

    while (self.cursor + size) > len(self.buf):
      remainder = self.buf[self.cursor:]
      res = res + remainder
      size = size - len(remainder)
      self.advanceCursor(len(remainder))
      if self.eof:
	return(res)
    
    # Buffer has enough material to fill the request:

    res = res + self.buf[self.cursor:self.cursor + size]
    self.advanceCursor(size)
    return(res)
      
  # Adv the buffer cursor. Remember that we have no control over how
  # much getEligibleLine will return to us. If dist is 0 and cursor
  # is at end of buffer, refresh buffer and set cursor to 0:

  def advanceCursor(self, dist):

    if (dist == 0) and (self.cursor >= len(self.buf)):
      self.buf = self.getEligibleLine()
      if self.verbosity:
	self.feedbackGadget.tick()
      if not self.buf:
	self.eof = 1
      self.cursor = 0
      return(1)

    # Get more info until we finally have enough to fill the advance request:
    while (self.cursor + dist) > len(self.buf):
      dist = dist - len(self.buf[self.cursor:])
      # Put more eligible material into the buffer:
      self.buf = self.getEligibleLine()
      if self.verbosity:
	self.feedbackGadget.tick()
      if not self.buf:
	self.eof = 1
	return(0)
      self.cursor = 0
    
    self.cursor = self.cursor + dist
    return(1)
      
  # Read from inFd till the next citation opener is found. Write out everything
  # we see that's not a citation opener. Buffer cursor is left right
  # after the citation opener. No return value.
  
  def readToNextCitation(self):

    while 1:
      chr = self.readBuf(1)
      if (chr == self.citationOpener[0]):
	window = chr
	while window:
	  possibleOpenerRest = self.readBuf(len(self.citationOpener) - len(window))
          if len(possibleOpenerRest) <> len(self.citationOpener) - len(window):
	    self.outFd.write(possibleOpenerRest)
	    return
	  window = window + possibleOpenerRest
	  if window == self.citationOpener:
	    return
	  self.outFd.write(window[0])
	  for i in range(len(window) - 1):
	    if window[i+1] <> self.citationOpener[0]:
	      self.outFd.write(window[i+1])
	      if i >= len(window) - 2:
		window = ''
	    else:
	      window = window[i+1:]
	      break
      elif not chr:
	return
      else:
	self.outFd.write(chr)
	
  # Helper for advanceCursor: this scans the file forward, writing to
  # the output, until it finds a string where a citation *could* be. For
  # random files, this method just returns the next line. But subclasses, such
  # as the MIF converter need to be smarter to scan past control info:
  
  def getEligibleLine(self):
    return(self.inFd.readline())


  # We assume that inFd pts to the start of one or more citations in the doc.
  # Return a list of them all. We scan for citationSeparator and citationCloser.
  # We leave the inFd just past the citationCloser:
  
  def getCiteKeyList(self):
    
    chr = ''
    citeKeyList = []
    while 1:       # Go through each citation key
      citeKey = ''
      while 1:     # Get a char at a time, looking for citation key delimiters
	# There might be a char already there from a previous cite key and
	# the corresponding look-ahead for this next one:
	if not chr:
	  chr = self.readBuf(1)
	if self.eof:
	  self.handleWarning("End of file in midst of citation.")
	  return(citeKeyList)
	if chr in self.citeKeyDelimiters:   
	  break
	else:
	  citeKey = citeKey + chr
	  chr = ''
      citeKeyList.append(citeKey)
      if chr == self.citationCloser:
        break
      else:
        # Find start of next cite key, or end of this set of citation keys:
        while 1:
  	  chr = self.readBuf(1)
	  if chr == self.citationCloser:
	    return(citeKeyList)
	  elif chr in self.citeKeyDelimiters:
	    continue
	  else:
	    # Found char of next citation in a citation key list:
	    break
      # Continue with inner loop for next citation
      continue
    return(citeKeyList)
  
    
  def getRefEntry(self, citeKey):
    entry = 0
    firstLuckyDb = 0
    for db in self.dbList:
      if db.has_key(citeKey):
	# Do we already have an entry (i.e. check for duplicates):
	if entry:
	  self.handleWarning("Key '" +\
			     citeKey +\
			     "' defined in db '" +\
			     firstLuckyDb.name +\
			     "' *and* in db '" +\
			     db.name +\
			     "'. Using entry in the former")
	else:
	  entry = db[citeKey]
	  firstLuckyDb = db

    return(entry)
    
  # If refListFileName is specified, put the reference list there as a
  # self-standing MIF file. Otherwise the reference list is appended to the
  # output file. This option is there because sometimes appending MIF files
  # does not work. (see bibmif script for a use of this).

  def transformDocument(
    self,
    inDocStr='',           # Works given a doc in a string,
    inFileName='',         # or the name of a file w/ the doc
    outFd=None,            # if provided, outFileName ignored (works only
                           # when bibl is inserted after a word (see above)
    systemFileHeader='',   # optional info to be placed just before the
                           # start of the final output file. This is NOT
                           # the application-specific one (like the MIF header)
                           # It's used for WWW support.
    outFileName = '',      # If empty and no outFd, return string w/ new doc.
    refListFileName='',    # If speced, put ref list there
    citationStyle='NumericCiteAppearanceOrder', # How citations & reflist  look inside the doc
    refStyle='GenericStyle', # How individual references will look
    insertionPt='',        # Word after which bibl is to be inserted
    annote=0,              # Have annotations included in bibliography?
    dbs = []):             # Bibliographic dbs to search
     
    # Make sure everything is reset:
    self.__init__(verbosity = self.verbosity)

    self.processParms(inDocStr,
		      inFileName,
		      outFileName,
		      outFd,
		      refListFileName,
		      citationStyle,
		      refStyle,
		      insertionPt,       # Word after which bibl is to be inserted
		      dbs,
		      systemFileHeader)
		      
    if self.verbosity:
      self.feedbackGadget = bibutils.fileConsumptionFeedbackGadget(self.inFd,
								   feedbackLeader = '... processed ',
								   feedbackTrailer = ' of document.',
								   reportPercentage = 1)
    # Get a reference generator object which takes entries and produces
    # platform-independent references, and a compiler object which takes
    # a list of references and produces a string with a reference list
    # for the target:
    self.gatherRefListMatters()

    # For citations with full authors, we force FullAuthorStyle.
    # It takes care of everything:
    if citationStyle == 'AuthorCiteAuthorAlphaOrder' or\
       citationStyle == 'AuthorCiteAppearanceOrder' or\
       ((type(citationStyle) == types.InstanceType) and\
	(citationStyle.__class__.__name__ == 'AuthorCiteAuthorAlphaOrder')) or\
       ((type(citationStyle) == types.InstanceType) and\
	(citationStyle.__class__.__name__ == 'AuthorCiteAppearanceOrder')):
      refStyle = 'FullAuthorStyle'

    self.prepareRefList(refStyle, annote=annote)
    # Compile it:
    res = self.compileAndOutput()
    self.cleanUp()
    return(res)

  def processParms(self,
		   inDocStr,
		   inFileName,
		   outFileName,
		   outFd,
		   refListFileName,
		   citationStyle,
		   refStyle,
		   insertionPt,      # Word after which bibl is to be inserted
		   dbs,
		   systemFileHeader):

    self.numberOfPasses = 1
    # Initialize the temp file name generation:
    tempfile.template = 'interbibTmp-'
    self.tmpName = tempfile.mktemp() + '.mif'
    self.openDocument(inFileName = inFileName, inDocStr = inDocStr)
    if outFd:
      self.finalProductOutFd = outFd
      # Need a local outfile for first pass:
      outFileName = tempfile.mktemp() + '.mif'
    if outFileName:
      self.outFd = bibutils.StrOrFileFd(fileName = outFileName, mode = 'w')
    else:
      self.outFd = bibutils.StrOrFileFd(str = self.outStr, mode = 'w')
    if refListFileName:
      self.refListFd = bibutils.StrOrFileFd(fileName = refListFileName, mode = 'w')
    else:
      self.refListFd = None

    # Was a word specified after which bibl is to be inserted?
    if insertionPt:
      self.numberOfPasses = self.numberOfPasses + 1
      self.bibliographyInsertPt = insertionPt
      self.insertionPtStr = insertionPt
      # Make sure the string user specified doesn't have any chars that are
      # special to regular expression searching. If it does, escape them:
      for specialChar in '$^.*+?[]\\':
	insertionPt = re.sub('\\' + specialChar, '\\' + specialChar ,insertionPt)
      # Make a compiled pattern matcher that's case insensitive:
      self.insertPtMatcher = re.compile(insertionPt, re.IGNORECASE)
    if not dbs:
      self.handleError("Method transformDocument requires a 'dbs' parameter",
		       TypeError, 'not enough arguments')
    self.dbList = dbs
    self.systemFileHeader = systemFileHeader

    # Make citation style:
    if type(citationStyle) == types.StringType:
      self.citationStyle = eval(citationStyle)()
      self.citationStyle.tab = self.tab

    if citationStyle == 'NumericCiteAuthorAlphaOrder':
      self.twoPasses = 1
      self.numberOfPasses = self.numberOfPasses + 1
      self.citationStartPointers = []
      self.finalOutFd = self.outFd
      self.outFd = bibutils.StrOrFileFd(fileName = self.tmpName, mode = 'w')
    else:
      self.twoPasses = 0
    
  # Go through file, find the citations, replace them with citations
  # according to style, and gather info needed to prepare the reference list.
  
  def gatherRefListMatters(self):
    while 1:

      # Scan and copy till the next (set of) citation(s) in the file:
      self.readToNextCitation()
      if self.eof:
	break  # Read the whole doc
      
      self.outFd.write(self.citationStyle.citationOpener)
      # If we'll have to do two passes through the file, remember where this
      # citation opens:
      if self.twoPasses:
	self.citationStartPointers.append(self.outFd.tell())
      
      # FD-pt is already past the citation opener:
      
      citeKeyList = self.getCiteKeyList()
      didACitation = 0

      # In the loop below: citeKey is what was found in the document body,
      # citationKey is what will replace citeKey in the body and will be
      # used to label the reference list:

      for citeKey in citeKeyList:
	citationKey = ''
	for i in range(len(self.refedEntries)):
	  if self.refedEntries[i][2] == citeKey:
	    # Referenced this entry before. Use key assigned then:
	    citationKey = self.refedEntries[i][1]
	if not citationKey:
	  entry = self.getRefEntry(citeKey)
          if not entry:
            self.handleMissingRef(citeKey)
	    citationKey = "?"
	  else:
 	    citationKey = self.citationStyle.citationKey(entry)
	  self.refedEntries.append([entry, citationKey, citeKey])
        if didACitation:
          self.outFd.write(self.citationStyle.citationSeparator)
        self.outFd.write(citationKey)
        didACitation = 1

      self.outFd.write(self.citationStyle.citationCloser)
    

  # Assumes that the following has been done: Read the doc, replaced all
  # citations with nice ones according to the citation style, and wrote the
  # whole thing to the outFd. Now make a list of
  # entry/platform-independent-ref pairs and do a second pass if numeric cite
  # with key/author alpha ordered reference list:

  def prepareRefList(self, refStyle, annote=0):

    refGen = bibrefgen.ReferenceGenerator()
    for entryCitationKeyCiteKeyTriple in self.refedEntries:
      entry = entryCitationKeyCiteKeyTriple[0]
      citationKey = entryCitationKeyCiteKeyTriple[1]
      # Entry might be null, if a citation key was not found!:
      if entry:

        self.refList.append([entry,
	  		     refGen.generateRef(entry, refStyle, annote=annote),
			     citationKey])


    # Pass the list of entry/generic-refs/citationKey triples to the
    # citation style for preparation of the properly sorted reference list
    # (with labels in front, etc.):

    # Get proper sorting done (in place):
    self.citationStyle.sort(self.refList)

    # If doing numeric cite, author-alpha order:
    # Go through the properly sorted entry/ref/citationKey triplets and build
    # a dictionary that maps citation key to the new position in the
    # reference list:
    if self.twoPasses:
      self.secondPass()

    # Prepare the final reference list:
    self.refList = self.citationStyle.prepareReferenceList(self.refList)


  def compileAndOutput(self):
    # Prepare formatting for the bibliography
    res = self.computeBibIntro()
    for ref in self.refList:
      res = res + self.refCompiler.compileRef(ref)

    if self.bibliographyInsertPt:
      if self.foundBibInsertPt:
        self.insertBibliography(res)
      else:
        self.handleError("Could not find insertion point for bibliography. " +\
			 "You specified insertion keyword '" +\
			 self.insertionPtStr +\
			 "'.", 
			 ValueError)
    elif self.refListFd:
      self.refListFd.write(self.fileHeader)
      self.refListFd.write(res)
      self.refListFd.close()
      self.outFd.close()
    else:
      # Finally, if bib wasn't supposed to be inserted after some word, reopen
      # file for updating and insert bibliography before any designated end of
      # file char (which is provided by each subclass):

      if self.outFd.isFile():
        fdName = self.outFd.name()
        self.outFd.close()
        self.outFd = bibutils.StrOrFileFd(fileName = fdName, mode = 'r+')

      if self.endFileChar:
	self.outFd.seek(-1,2)
	while 1:
	  chr = self.outFd.read(1)
	  if chr != self.endFileChar:
	    if self.outFd.tell() <= 1:
	      raise ValueError, "Could not find closing '" +\
		    self.endFileChar +\
		    "' when adding bibliography to end."
	    self.outFd.seek(-2,1)
	    continue
	  else:
	    self.outFd.seek(-1,1)
	    break
	self.outFd.write(res)
	self.outFd.write(self.endFileChar)
      else:
	self.outFd.seek(0,2)
	self.outFd.write(res)
      self.outFd.close()

    if self.finalProductOutFd or self.outFd.name():
      return(1)
    else:
      # Since string writing isn't in-place, need to get the output string out of the
      # output fd:
      self.outStr = self.outFd.str
      return(self.outStr)

  def secondPass(self):

    citationMap = {}
    for i in range(len(self.refList)):
      citationMap[self.refList[i][2]] = str(i+1)

    # Do the second pass:
    self.outFd.close()
    self.outFd = self.finalOutFd
    # Reopen the file we just wrote the partial final MIF doc into:
    fd = open(self.tmpName, 'r')
    if self.verbosity:
      self.feedbackGadget = bibutils.fileConsumptionFeedbackGadget(fd,
								   feedbackLeader = '... Pass 2 of ' + \
								   str(self.numberOfPasses) + \
								   ': processed ',
								   feedbackTrailer = ' of document.',
								   reportPercentage = 1)

    for citationStartPt in self.citationStartPointers:
      if self.verbosity:
        self.feedbackGadget.tick()
      filePiece = fd.read(citationStartPt-fd.tell())
      # Write everything including the citation opener:
      self.outFd.write(filePiece)
      if self.verbosity:
        self.feedbackGadget.tick()
      self.replaceCitation(fd, citationMap)
    # All citation are replaced. Copy the rest of the file over:
    rest = fd.read()
    fd.close()
    self.outFd.write(rest)

  # We expect the fd's cursor to be just past a citation opening with
  # numeric style: [2, 24]. CitationMap is a dict mapping original citation
  # number to final number. We read the citations and write a corresponding
  # list, but w/ new cite numbers. Note that there may be a line break
  # anywhere within the citation. We need to handle that:

  def replaceCitation(self, fd, citationMap):
    done = 0
    while self.readToNextCitationOrCitationEnd(fd):
      citation = bibutils.readNum(fd)

      self.outFd.write(citationMap[citation])

  def readToNextCitationOrCitationEnd(self, fd):
    while 1:
      chr = fd.read(1)
      if chr == '': # EOF
	return(0)
      if bibutils.digitsOnly(chr):
	# Found another citation:
	fd.seek(-1, 1)
	return(1)
      self.outFd.write(chr)
      if chr == self.citationStyle.citationCloser:
	return(0)

  def cleanUp(self):
    if os.path.isfile(self.tmpName):
      os.system('rm ' + self.tmpName)
    if os.path.isfile(self.macFileTmpName):
      os.system('rm ' + self.macFileTmpName)

  def setDocumentCitationDelimiters(self, citationOpener, citationCloser):
    self.citationOpener = citationOpener
    self.citationCloser = citationCloser

class MIFDocTransformer(DocTransformer):

  mifStrOpen = 0
  mifBuf     = ''
  mifPos     = 0

  # NOTE: This init method is called multiple times!!
  def __init__(self, verbosity = 1):
    DocTransformer.__init__(self, verbosity)
    self.mifStrOpen = 0
    self.mifPos = 0
    self.mifBuf = ''
    self.fileHeader = "<MIFFile 5.00> # Generated by InterBib\n"
    self.refCompiler = bibcompile.BibMIFCompiler()
    self.endFileChar = ''

  # Helper for advanceCursor: this scans the file forward, writing to
  # the output, until it finds a string where a citation *could* be. For
  # MIF files, these are strings enclosed within '<String `...'>'. 
  # If self.mifStrOpen is true, the fd is in the middle of one of those
  # strings. In that case we look for the close and return that piece. Else
  # we look of the opening and then pass at least some of the string up:
  
  def getEligibleLine(self):

    # If the last call read in a MIF string, that is all written out now,
    # but not the closing: "'>".

    if self.mifStrOpen:

      # self.mifPos was left at the closing '>' of a str. But the closing
      # chars "'>" were not written yet. Write them and get ready for the
      # search onward:

      self.outFd.write("'>")
      # Pt past the closing angle bracket:
      self.mifPos = self.mifPos + 1
      self.mifStrOpen = 0

    while 1:

      if self.mifPos >= len(self.mifBuf):
        if not self.readIntoMIFBuffer():
	  return('')

      # If the line starts with an '&', it's binary data, don't look at it:

      if self.mifBuf[0] == '&':
	self.outFd.write(self.mifBuf)
	self.mifPos = len(self.mifBuf)
        continue

      strStart = string.find(self.mifBuf, "<String `", self.mifPos)

      if strStart > -1:

	# Write everything till where the string actually starts:
	self.outFd.write(self.mifBuf[:strStart + len("<String '")])

	# Now scan along till we find the closing '>':

	lookedAtBackslash = 0
	realStrStart = strStart + len("<String '")
	self.mifPos = realStrStart

	while 1:

	  if self.mifPos >= len(self.mifBuf):
            if not self.readIntoMIFBuffer():
	      return('')
	  chr = self.mifBuf[self.mifPos]
	  self.mifPos = self.mifPos + 1
	  if lookedAtBackslash:
	    lookedAtBackslash = 0
	    continue
	  # start of a backslashed something?
	  if chr == '\\':
	    # Make sure we ignore the following char:
	    lookedAtBackslash = 1
	    continue
	  # Closing the string?:
	  if chr == "'":
	    # We'll return everything from after "<String `" to before the closing
	    # "'":
	    res = self.mifBuf[realStrStart:self.mifPos-1]
	    break
        self.mifStrOpen = 1
        # self.mifPos is now pointing to the closing '>'
        return(res)

      else:
	# It was not a line with a MIF string directive in it. Keep looking:
	self.outFd.write(self.mifBuf[self.mifPos:])
	self.mifPos = len(self.mifBuf)
	continue

  def readIntoMIFBuffer(self):

    self.mifBuf = self.inFd.readline()
    self.mifPos = 0
    if self.verbosity:
      self.feedbackGadget.tick()
    if not self.mifBuf:
      self.eof = 1
      return(0)
    else:
      # If we are supposed to stick the bibliography somewhere, look for
      # the spot and remember where it is in the outfile (look all the
      # time, because we'll use the *last* occurrence of the word):
      if self.bibliographyInsertPt:
	try:
	  refMatchObj = self.insertPtMatcher.search(self.mifBuf)
	  insertWordIndex = refMatchObj.start()
	  self.bibliographyInsertPt = self.outFd.tell() + insertWordIndex
	  self.foundBibInsertPt = 1
        except:
	  pass
      return(1)
       
  # Insert bibliography in MIF file after word starting at position
  # self.bibliographyInsertPt. We assume that the second pass has
  # already been done. We mv the outfile to the tempfile. Then we
  # copy the out file up to the insert pt. Then we find the paragraph
  # closing bracket and insert the bibl:
  
  def insertBibliography(self, bibl):
    outFileName = self.outFd.name()
    # If this wasn't a string file, need to do some file moving:
    if outFileName:
      self.outFd.close()
      os.system('mv ' + outFileName + ' ' + self.tmpName)
      self.inFd = bibutils.StrOrFileFd(fileName=self.tmpName)
    else:
      self.outFd.seek(0,0)
      self.inFd = self.outFd
    if self.finalProductOutFd:
      self.outFd = self.finalProductOutFd
    else:
      self.outFd = bibutils.StrOrFileFd(fileName = outFileName, mode = 'w')
    # Output the optional file header (NOT the appl.-specific one, like
    # the MIF header. This is just for WWW support):
    if self.systemFileHeader:
      self.outFd.write(self.systemFileHeader)
    if self.verbosity:
      self.feedbackGadget = bibutils.fileConsumptionFeedbackGadget(self.inFd,
								   feedbackLeader = '... Pass 3 of ' + str(self.numberOfPasses) + ': processed ',
								   feedbackTrailer = ' of document.',
								   reportPercentage = 1)


    if self.verbosity:
      self.feedbackGadget.tick()
    # Read to start of word after which bibl is to be inserted:
    buf = self.inFd.read(self.bibliographyInsertPt)
    if self.verbosity:
      self.feedbackGadget.tick()
    self.outFd.write(buf)
    # Now look for the closing Para angle bracket:
    lookedAtBackslash = 0
    # At least string, paraline and para are open right now:
    openBlock = 3
    while 1:
      chr = self.inFd.read(1)
      if not chr:
	self.handleError("Premature EOF during bibl insertion")
	return(-1)
      self.outFd.write(chr)
      if lookedAtBackslash:
	lookedAtBackslash = 0
	continue
      # start of a backslashed something?
      if chr == '\\':
	# Make sure we ignore the following char:
	lookedAtBackslash = 1
	continue
      # Opening a new block?:
      if chr == "<":
	openBlock = openBlock + 1
      # Closing a block?:
      if chr == ">":
	openBlock = openBlock - 1
	if openBlock == 0:
	  break
    # Found the paragraph closing after the word where bibl is to be inserted.
    # Make the splice-in look nice:
    self.outFd.write('\n ')
    self.outFd.write(bibl)
    # Xfer the rest of the MIF file:
    buf = self.inFd.read()
    self.outFd.write(buf)
    # Only close out fd if it wasn't one specified by the caller:
    if not self.finalProductOutFd:
      self.outFd.close()
    self.inFd.close()

  def computeBibIntro(self):
    # Start with setting the part format to 'Body':
    return "<Para\n  <PgfTag `Body'>\n > # end of Para\n"


class RTFDocTransformer(DocTransformer):

  rtfExtra = None
  rtfBuf   = ''
  rtfPos   = 0

  # NOTE: This init method is called multiple times!!
  def __init__(self, verbosity = 1):
    DocTransformer.__init__(self, verbosity)
    self.rtfExtra = None
    self.rtfPos = 0
    self.rtfBuf = ''
    self.fileHeader = "{\rtf1\ansi "
    self.refCompiler = bibcompile.BibRTFCompiler()
    self.endFileChar = '}'


  # Utility for getEligibleLine - method writes out remainder of group
  # (delineated by brackets) to output file.  Used to skip groups that
  # might have non-body text (i.e. font names, etc)

  def skipRTFGroup(self):

    level = 1               # one open bracket already encountered
    startPos = self.rtfPos  # keep track of start of string

    while level > 0:

      if self.rtfPos >= len(self.rtfBuf):
        self.outFd.write(self.rtfBuf[startPos:self.rtfPos])  # write out line
        startPos = 0
        if not self.readIntoRTFBuffer():
          return -1                                       # error in RTF file!

      if self.rtfBuf[self.rtfPos] == '{':
        level = level + 1
      elif self.rtfBuf[self.rtfPos] == '}':
        level = level - 1

      self.rtfPos = self.rtfPos + 1

    self.outFd.write(self.rtfBuf[startPos:self.rtfPos])   # write out group


  # Utility for getEligibleLine - Reads in RTF control word and returns
  # tuple of (control word, parameter).

  def skipRTFControlWord(self):

    startPos = self.rtfPos    # Position should be '\'

    if self.rtfPos >= len(self.rtfBuf): # check for newline
      if not self.readIntoRTFBuffer():
        return -1                       # This is an error, should end with }
      return "\\\n"

    # If second character is a ', this is an ASCII code

    if self.rtfBuf[self.rtfPos+1] == "'":
      self.rtfPos = self.rtfPos + 4
      return (self.rtfBuf[startPos:self.rtfPos], -1)

    # If second character is not a lower-case letter, this is a control
    # symbol

    if self.rtfBuf[self.rtfPos+1] not in string.lowercase:
      self.rtfPos = self.rtfPos + 2
      return (self.rtfBuf[startPos:self.rtfPos], -1)

    # Otherwise this is a control word, search until delimiter is found

    self.rtfPos = self.rtfPos + 2  # we know the second character from above

    while 1:

      if self.rtfBuf[self.rtfPos] in string.lowercase:
        self.rtfPos = self.rtfPos + 1

      # Blank delimiter - return word w/o blank, increment rtfPos past blank

      elif self.rtfBuf[self.rtfPos] == ' ':
        res = self.rtfBuf[startPos:self.rtfPos]
        self.rtfPos = self.rtfPos + 1
        return (res, -1)

      # Numeric parameter - read in the parameter

      elif self.rtfBuf[self.rtfPos] in string.digits + '-':
        endPos = self.rtfPos
        res = self.rtfBuf[startPos:endPos]
        self.rtfPos = self.rtfPos + 1
        while self.rtfBuf[self.rtfPos] in string.digits:
          self.rtfPos = self.rtfPos + 1
        param = string.atoi(self.rtfBuf[endPos:self.rtfPos])
        if self.rtfBuf[self.rtfPos] == ' ':
          self.rtfPos = self.rtfPos + 1
        return (res, param)

      # Symbolic delimiter - part of next item so leave rtfPos on it

      else:
        res = self.rtfBuf[startPos:self.rtfPos]
        return (res, -1)


  # Utility for skipRTFObject - skips over binary data and resets buffer.
  # Length of binary field held in parameter lngth.

  def skipRTFBinary(self, lngth):
    if lngth < len(self.rtfBuf) - self.rtfPos:
      self.outFd.write(self.rtfBuf[self.rtfPos:self.rtfPos+lngth])
      self.rtfPos = self.rtfPos + lngth
    else:
      self.outFd.write(self.rtfBuf[self.rtfPos:])
      i = lngth - (len(self.rtfBuf) - self.rtfPos)
    lngLine = self.inFd.read(i)
    self.outFd.write(lngLine)
    if not self.readIntoRTFBuffer():
      return -1                          # error in RTF file!


  # Utility for skipRTFObject - skips over hex data and resets buffer.
  # Hex data should end with the '}' symbol.

  def skipRTFHex(self):
    res = string.find(self.rtfBuf, "}", self.rtfPos)
    if res != -1:
      self.outFd.write(self.rtfBuf[self.rtfPos:res+1])
      self.rtfPos = res + 1
      return
    self.outFd.write(self.rtfBuf[self.rtfPos:])

    while res == -1:                     # Search by line for '}'
      if not self.readIntoRTFBuffer():
        return -1                        # error in RTF file!
      res = string.find(self.rtfBuf, "}")
      if res != -1:
         self.rtfPos = res + 1
         self.outFd.write(self.rtfBuf[0:self.rtfPos])
      else:
         self.outFd.write(self.rtfBuf)


  # Utility for getEligibleLine - method writes out group containing
  # binary or hexidecimal data, i.e. pictures and objects.

  def skipRTFObject(self):

    # Search through the control words at the front of the group.  If
    # the \bin control word is encountered, data is binary.  Otherwise,
    # data is in hex

    while 1:

      if self.rtfPos >= len(self.rtfBuf):
        if not self.readIntoRTFBuffer():
          return('')

      if self.rtfBuf[self.rtfPos] == '\\':
        startCntl = self.rtfPos
        word, param = self.skipRTFControlWord()
        self.outFd.write(self.rtfBuf[startCntl:self.rtfPos])
        if word == '\\bin':
          self.skipRTFBinary(param)      # binary data
          return
      else:
        self.skipRTFHex()                # hex data
        return


  # Helper for advanceCursor: this scans the file forward, writing to
  # the output, until it finds a string of text that a citation could be in.

  def getEligibleLine(self):

    # Data:
    #   escChars - List of control words and characters that count as text
    #   skipWords - Words that precede a long group that can safely be skipped
    #   objWords - Words that precede a group containing hex/binary data

    escChars = ['\\', '\{', '\}', '\bullet', '\endash', '\emdash', '\lquote',
                '\rquote', '\ldblquote', '\rdblquote']
    skipWords = ['\fonttbl', '\colortbl', '\filetbl', '\stylesheet', '\revtbl',
                 '\info', '\bkmkstart', '\bkmkend', '\atnauthor', '\atnid',
                 '\field']
    objWords = ['\objdata', '\objalias', '\objsect', '\pict']

    # variable refflag is 1 whenever we're in the middle of a reference

    refflag = 0


    # If some additional text was read in the last call to this method,
    # write it out now
  
    if self.rtfExtra != None:
      self.outFd.write(self.rtfExtra)
      self.rtfExtra = None

    while 1:   # First loop -- find beginning of text

      if self.rtfPos >= len(self.rtfBuf):
        if not self.readIntoRTFBuffer():
          return('')

      # If current character is '\', get control word.  If it is an escaped
      # text character or an ASCII code, move to second loop; else check if
      # the word is the start of a group or object to skip.

      if self.rtfBuf[self.rtfPos] == '\\':
        strStart = self.rtfPos
        word, param = self.skipRTFControlWord()
        if word in escChars or word[0:2] == "\\'":
          break
        else:
          self.outFd.write(self.rtfBuf[strStart:self.rtfPos])
          if word in skipWords:
            self.skipRTFGroup()
          elif word in objWords:
            self.skipRTFObject()

      # Else if the current character is a bracket, keep searching

      elif self.rtfBuf[self.rtfPos] in ['{', '}']:
        self.outFd.write(self.rtfBuf[self.rtfPos])
        self.rtfPos = self.rtfPos + 1

      # Else character is the start of text, so move to the second loop

      else:
        strStart = self.rtfPos
        self.rtfPos = self.rtfPos + 1
        if self.rtfBuf[strStart] == '[':
          refflag = 1
        break


    while 1:   # Second loop -- find end of text

      # If EOL is found, check if we are in the middle of a reference.  If
      # so, read in a new line and continue, otherwise return

      if self.rtfPos >= len(self.rtfBuf):

        if refflag:
          tempbuf = self.rtfBuf[strStart:self.rtfPos]
          if not self.readIntoRTFBuffer():
            return tempbuf
          else:
            # Note - if ref takes more than two lines, InterBib will "burp"
            tempbuf = tempbuf[:-1]
            res = string.find(self.rtfBuf, ']')
            self.rtfPos = res + 1
            return tempbuf + self.rtfBuf[0:self.rtfPos]
        else:
          return self.rtfBuf[strStart:self.rtfPos]

      # If current character in buffer is '\', get control word.  If it
      # is an escaped text character, continue; otherwise save the control
      # word in rtfExtra and return the text

      if self.rtfBuf[self.rtfPos] == '\\':
        strEnd = self.rtfPos
        word, param = self.skipRTFControlWord()
        if word not in escChars and word[0:2] != "\\'":   # found end of text
          res = self.rtfBuf[strStart:strEnd]
          self.rtfExtra = self.rtfBuf[strEnd:self.rtfPos]
          return res

      # Else if current character is a squiggly bracket, save it in rtfExtra
      # and return the text

      elif self.rtfBuf[self.rtfPos] in ['{', '}']:
        res = self.rtfBuf[strStart:self.rtfPos]
        self.rtfExtra = self.rtfBuf[self.rtfPos]
        self.rtfPos = self.rtfPos + 1
        return res

      # Else the current character is text, keep searching

      else:
        if self.rtfBuf[self.rtfPos] == '[':
          refflag = 1
        elif self.rtfBuf[self.rtfPos] == ']':
          refflag = 0
        self.rtfPos = self.rtfPos + 1


  def readIntoRTFBuffer(self):
    """Reads next line from RTF file into buffer"""
    self.rtfBuf = self.inFd.readline()
    self.rtfPos = 0
    if self.verbosity:
      self.feedbackGadget.tick()
    if not self.rtfBuf:    # EOF, return 0
      self.eof = 1
      return 0
    else:
      # If we are supposed to stick the bibliography somewhere, look for
      # the spot and remember where it is in the outfile (look all the
      # time, because we'll use the *last* occurrence of the word):
      if self.bibliographyInsertPt:
	try:
	  refMatchObj = self.insertPtMatcher.search(self.rtfBuf)
	  insertWordIndex = refMatchObj.start()
	  self.bibliographyInsertPt = self.outFd.tell() + insertWordIndex
	  self.foundBibInsertPt = 1
        except:
	  pass
      return(1)

  # Insert bibliography in RTF file after word starting at position
  # self.bibliographyInsertPt. We assume that the second pass has
  # already been done. We mv the outfile to the tempfile. Then we
  # copy the out file up to the insert pt. Then we find the paragraph
  # closing bracket and insert the bibl:
  
  def insertBibliography(self, bibl):
    outFileName = self.outFd.name()
    # If this wasn't a string file, need to do some file moving:
    if outFileName:
      self.outFd.close()
      os.system('mv ' + outFileName + ' ' + self.tmpName)
      self.inFd = bibutils.StrOrFileFd(fileName=self.tmpName)
    else:
      self.outFd.seek(0,0)
      self.inFd = self.outFd
    if self.finalProductOutFd:
      self.outFd = self.finalProductOutFd
    else:
      self.outFd = bibutils.StrOrFileFd(fileName = outFileName, mode = 'w')
    # Output the optional file header (NOT the appl.-specific one, like
    # the MIF header. This is just for WWW support):
    if self.systemFileHeader:
      self.outFd.write(self.systemFileHeader)
    if self.verbosity:
      self.feedbackGadget = bibutils.fileConsumptionFeedbackGadget(self.inFd,
								   feedbackLeader = '... Pass 3 of ' + str(self.numberOfPasses) + ': processed ',
								   feedbackTrailer = ' of document.',
								   reportPercentage = 1)

    if self.verbosity:
      self.feedbackGadget.tick()
    # Read to start of word after which bibl is to be inserted:
    buf = self.inFd.read(self.bibliographyInsertPt + len(self.insertionPtStr))
    if self.verbosity:
      self.feedbackGadget.tick()
    self.outFd.write(buf)

    # Look for first word that is not a backslashed control, or for a
    # closing '}':
    lookedAtBackslash = 0
    foundClosingBrace = 0
    while 1:
      chr = self.inFd.read(1)
      if not chr:
	self.handleError("Premature EOF during bibl insertion")
	return(-1)
      if lookedAtBackslash:
	if chr in ['\n', ' ']:
  	  lookedAtBackslash = 0
	self.outFd.write(chr)
	continue
      # start of a backslashed something?
      if chr == '\\':
	# Make sure we ignore the following chars:
	lookedAtBackslash = 1
	self.outFd.write(chr)
	continue
      # Closing block?:
      if chr == "}":
	foundClosingBrace = 1
        break
      self.outFd.write(chr)
      break

    # Found the paragraph closing after the word where bibl is to be inserted.
    # Make the splice-in look nice:
    self.outFd.write('\\par\n ')
    self.outFd.write(bibl)
    if foundClosingBrace:
      self.outFd.write('}')
    # Xfer the rest of the RTF file:
    buf = self.inFd.read()
    self.outFd.write(buf)
    # Only close out fd if it wasn't one specified by the caller:
    if not self.finalProductOutFd:
      self.outFd.close()
    self.inFd.close()

  def computeBibIntro(self):
    # Start with setting the part format to 'Body':
    return "\\par "

class HTMLDocTransformer(DocTransformer):

  # NOTE: This init method is called multiple times!!
  def __init__(self, verbosity = 1):
    DocTransformer.__init__(self, verbosity)
    self.htmlBuf = ''
    self.bufPt = -2 # Nothing read in yet
    self.refCompiler = bibcompile.BibHTMLCompiler()
    self.endFileChar = ''
    self.tab = '\x09'

  # Helper for advanceCursor: this scans the file forward, writing to
  # the output, until it finds a string where a citation *could* be. For
  # HTML files, these are strings that have no tags.
  # If self.mifStrOpen is true, the fd is in the middle of one of those
  # strings. In that case we look for the close and return that piece. Else
  # we look of the opening and then pass at least some of the string up:
  
  def getEligibleLine(self):

    # If this is the first call, pull the whole html file into a buffer.
    # In the MIF and RTF transformers we are fancier and do this in
    # smaller chunks. Here we trust that the files are not so outrageous
    # that they'll swap like crazy.
    # Note, however, that feedback gadgets key off the file pointer. This
    # means that user feedback will always see the process as '100% complete'.
    # So we need to adjust the file pointer whenever we adjust the buffer
    # pointer:

    if self.bufPt == -2:
      self.htmlBuf = self.inFd.read()
      self.bufPt = 0
      # If we are supposed to stick the bibliography somewhere, look for
      # the spot and remember where it is in the outfile (i.e. find the
      # last occurrence of the bibl marker word). This is only approximate
      # because the file length of the output is different from this
      # input. We are just interested in whether the word is there:
      if self.bibliographyInsertPt:
	insertWordIndex = string.rfind(self.htmlBuf, self.insertionPtStr)
	if insertWordIndex > -1:
	  self.foundBibInsertPt = 1
	  self.bibliographyInsertPt = insertWordIndex

    # If self.bufPt is -1, we are at the end of the file:
    if self.bufPt == -1:
      # Set in file pt to end of in file:
      self.inFd.seek(0,2)
      return('')

    # In the last call, did we read up to a tag? If yes, skip over
    # it now (writing it to output):
    while self.htmlBuf[self.bufPt] == '<':
      tagEnd = string.find(self.htmlBuf, '>', self.bufPt)
      if tagEnd == -1:
	raise ValueError, 'Could not find end of tag ' +\
	      self.htmlBuf[self.bufPt:self.bufPt+60] + '...'
      self.outFd.write(self.htmlBuf[self.bufPt:tagEnd+1])
      self.bufPt = tagEnd + 1
      # Check for end of buffer (occurs when doc ends with a '>'):
      if self.bufPt >= len(self.htmlBuf):
        break
    
    oldbufPt = self.bufPt
    # Return everything up to the next tag:
    self.bufPt = string.find(self.htmlBuf, '<', self.bufPt)
    # Update file pt in inFile to correspond to the buffer pt (to get user
    # feedback for percent completion right). Then return the info:
    if self.bufPt > -1:
      self.inFd.seek(self.bufPt,0)
      return(self.htmlBuf[oldbufPt:self.bufPt])
    else:
      # Reached end of input buffer. Indicate 'entire file read' by pting to EOF:
      self.inFd.seek(0,2)
      return(self.htmlBuf[oldbufPt:])

  # Insert bibliography in HTML file after word starting at position
  # self.bibliographyInsertPt. We assume that the second pass has
  # already been done. We mv the outfile to the tempfile. Then we
  # copy the out file up to the insert pt. Then we just add the bibl.
  
  def insertBibliography(self, bibl):
    outFileName = self.outFd.name()
    # If this wasn't a string file, need to do some file moving:
    if outFileName:
      self.outFd.close()
      os.system('mv ' + outFileName + ' ' + self.tmpName)
      self.inFd = bibutils.StrOrFileFd(fileName=self.tmpName)
    else:
      self.outFd.seek(0,0)
      self.inFd = self.outFd
    if self.finalProductOutFd:
      self.outFd = self.finalProductOutFd
    else:
      self.outFd = bibutils.StrOrFileFd(fileName = outFileName, mode = 'w')
    # Output the optional file header (NOT the appl.-specific one, like
    # the MIF header. This is just for WWW support):
    if self.systemFileHeader:
      self.outFd.write(self.systemFileHeader)
    if self.verbosity:
      self.feedbackGadget = bibutils.fileConsumptionFeedbackGadget(self.inFd,
								   feedbackLeader = '... Pass 3 of ' + str(self.numberOfPasses) + ': processed ',
								   feedbackTrailer = ' of document.',
								   reportPercentage = 1)


    if self.verbosity:
      self.feedbackGadget.tick()
    # Read the whole file:
    buf = self.inFd.read()
    if self.verbosity:
      self.feedbackGadget.tick()
    insertWordIndex = string.rfind(buf, self.insertionPtStr)
    # Now go past the word and past any HTML tags after it (often
    # this will be a section heading) and past any whitespace:
    # Pt past the insertion word:

    bufPt = insertWordIndex + len(self.insertionPtStr)

    htmlEndMatcher = re.compile('</html>', re.IGNORECASE)
    try:
      refHTMLEndObj = htmlEndMatcher.search(buf)
      endOfText = refHTMLEndObj.start()
    except:
      endOfText = None

    # Skip past whitespace and tags after the insertion place word:
    while (bufPt < len(buf)):
      if buf[bufPt] in string.whitespace:
	bufPt = bufPt + 1
	continue
      # Pointing to the ending tags or EOF?:
      if bufPt == endOfText:
	break
      # Pointing to a tag?
      if buf[bufPt] == '<':
	# Skip past tag(s):
	while 1:
	  # Read a piece of tag or the '>':
	  bufPt = bufPt + 1
	  try:
	    if buf[bufPt] == '>':
	      # Pt past the tag
	      bufPt = bufPt + 1
	      break
	  except:
	    raise ValueError, 'Could not find end of tag after last occurrence of ' +\
		  self.insertionPtStr
      else:
	# Found a non-whitespace/non-tag substring. Pt before it:
	bufPt = bufPt - 1
	break

    self.outFd.write(buf[:bufPt])
    self.outFd.write(bibl)
    # Xfer the rest of the HTML file:
    self.outFd.write(buf[bufPt:])
    # Only close out fd if it wasn't one specified by the caller:
    if not self.finalProductOutFd:
      self.outFd.close()
    self.inFd.close()

  def computeBibIntro(self):
    return "\n"

class CitationStyle:
  
  def __init__(self):
    self.citationSeparator = ', '
    self.citationOpener = '['       # delimiters for citations in final text
    self.citationCloser = ']'
    self.refListItemDelimiterOpen = '['
    self.refListItemDelimiterClose = ']'
    self.tab = bibconvert.FrameTAB

  # Given a list of [entry, <reference>, citationKey] triples, return a
  # sorted list of references only, such that they have leading labels, and
  # everything:
  
  def prepareReferenceList(self, entryRefKeyTripleList):

    finalRefList = []
    for i in range(len(entryRefKeyTripleList)):
      entryRefKeyTriple = entryRefKeyTripleList[i]
      refObj = entryRefKeyTriple[1]
      if self.__class__.__name__ == 'NumericCiteAuthorAlphaOrder':
	citationKey = str(i+1)
      else:
        citationKey = entryRefKeyTriple[2]
      # Get a label for the reference entry:
      refObj.prependTstr(bibrefgen.TaggedStr(bibrefgen.plain,
					     self.refListItemDelimiterOpen +\
					     citationKey +\
					     self.refListItemDelimiterClose +\
					     self.tab))
      finalRefList.append(refObj)
    return(finalRefList)


class NumericCiteKey(CitationStyle):
  citeKeyFormat = 'numeric'
  prevCiteKey   = 0

  # Given an entry, return a string for the citation in the document body:
  def citationKey(self, entry):
    self.prevCiteKey = self.prevCiteKey + 1
    return(str(self.prevCiteKey))

class AlphaCiteKey(CitationStyle):
  citeKeyFormat = 'alpha'
  
  # Given an entry, return a string for the citation in the document body:
  def citationKey(self, entry):
    return(entry.key)

class AuthorCiteKey(CitationStyle):
  citeKeyFormat = 'author'
  
  def __init__(self):
    self.citationSeparator = '; '
    self.citationOpener = '('     # delimiters for citations in final text
    self.citationCloser = ')'
    self.refListItemDelimiterOpen = '' # delimiters in reflist
    self.refListItemDelimiterClose = ''
    self.citeMemory = {}

  # Given an entry, return a string for the citation in the document body:
  def citationKey(self, entry):
    if not bibutils.getEntryFldSafely(entry, 'author'):
      if bibutils.getEntryFldSafely(entry, 'title'):
	key = entry.title
      else:
	key = entry.key
    else:
      fullNameAuthors = string.split(entry.author, bibimport.authorSeparator)
      lastNameAuthors = []
      for author in fullNameAuthors:
        author = string.split(bibutils.getAuthorLastNameFirst(author), ',')
        lastNameAuthors.append(author[0])
      if len(lastNameAuthors) > 3:
        key = lastNameAuthors[0] + ' et al.'
      elif len(lastNameAuthors) == 3:
	key = lastNameAuthors[0] +\
	      ', ' +\
	      lastNameAuthors[1] +\
	      ', and ' +\
	      lastNameAuthors[2]
      elif len(lastNameAuthors) == 2:
	key = lastNameAuthors[0] +\
	      ' and ' +\
	      lastNameAuthors[1]
      else: # single author
	key = lastNameAuthors[0]

    realYear = bibutils.getEntryFldSafely(entry, 'year')
    year = realYear
    if self.citeMemory.has_key(key + ', ' + realYear):
      prevOccurrenceYear = self.citeMemory[key + ', ' + realYear]
      year = self.uniquifyYear(prevOccurrenceYear)
      # This attribute will be deleted later by bibrefgen. But it needs
      # it to format the entry properly:
      entry.yearKey = year
    self.citeMemory[key + ', ' + realYear] = year
    if year:
      key = key + ', ' + year
    return(key)

  def uniquifyYear(self, existingYear):
    if bibutils.digitsOnly(existingYear):
      return(existingYear + 'a')
    lastChrAscii = ord(existingYear[-1])
    newCharAscii = lastChrAscii + 1
    if newCharAscii > ord('z'):
      raise ValueError, 'Too many ambiguous entries'
    return(existingYear[:-1] + chr(newCharAscii))


class AppearanceOrder(CitationStyle):

  # Sorts list of [entry, <reference>, citationKey] triples in place by
  # order of appearance in document body. Since it's in order already, do nothing:

  def sort(self, entryRefKeyTripleList):
    pass

class KeyAlphaOrder(CitationStyle):

  # Sorts list of [entry, <reference>, citationKey] triples in place by
  # alpha order of citation keys (

  def sort(self, entryRefKeyTripleList):
    entryRefKeyTripleList.sort(keySort)

class AuthorAlphaOrder(CitationStyle):

  # Sorts list of [entry, <reference>, citationKey] triples in place by
  # alpha order of authors:

  def sort(self, entryRefKeyTripleList):
    entryRefKeyTripleList.sort(authorSort)

# Here come the mixins that are used by clients:

class NumericCiteAppearanceOrder(NumericCiteKey, AppearanceOrder, CitationStyle):
  pass

class NumericCiteAuthorAlphaOrder(NumericCiteKey, AuthorAlphaOrder, CitationStyle):
  # NOTE: This requires two passes through the doc
  pass

class AlphaCiteKeyAlphaOrder(AlphaCiteKey, KeyAlphaOrder, CitationStyle):
  pass

class AlphaCiteAuthorAlphaOrder(AlphaCiteKey, AuthorAlphaOrder, CitationStyle):
  pass

class AlphaCiteAppearanceOrder(AlphaCiteKey, AppearanceOrder, CitationStyle):
  pass
     
class AuthorCiteAuthorAlphaOrder(AuthorCiteKey, AuthorAlphaOrder, CitationStyle):

  # The ref list preparation is very simple because no tags are added
  # before each entry. So just take each triplet, pull out the 2nd
  # entry which is the reference, and return a list of all of them:
  
  def prepareReferenceList(self, entryRefKeyTripleList):
    return(map(lambda x: x[1], entryRefKeyTripleList))

class AuthorCiteAppearanceOrder(AuthorCiteKey, AppearanceOrder, CitationStyle):

  # The ref list preparation is very simple because no tags are added
  # before each entry. So just take each triplet, pull out the 2nd
  # entry which is the reference, and return a list of all of them:
  
  def prepareReferenceList(self, entryRefKeyTripleList):
    return(map(lambda x: x[1], entryRefKeyTripleList))
  
# Called with two lists: [entry, <ref>, citationKey]. We only care about
# the keys in the entry parts. Return -1 if key1 < key2, return 0 if key1
# == key2, return 1 if key1 > key2:

def keySort(list1, list2):

  citationKey1 = list1[2]
  citationKey2 = list2[2]

  if bibutils.digitsOnly(citationKey1) and bibutils.digitsOnly(citationKey2):
    citationKey1 = string.atoi(citationKey1)
    citationKey2 = string.atoi(citationKey2)

  if citationKey1 < citationKey2:
    return(-1)
  if citationKey1 == citationKey2:
    return(0)
  else:
    return(1)
  
# Called with two lists: [entry, <ref>, citationKey]. We only care about
# the entry parts. Return -1 if author(entry1) < author(entry2), return 0
# if author(entry1) == author(entry2), or return 1 if author(entry1) >
# author(entry2):

def authorSort(list1, list2):
  return(bibutils.entryNameComparison(list1[0],
				      list2[0]))
