################################################################################
#
# File:         bibimport.py
# RCS:          $Header: /u/testbed/CVSROOT/dldev/src/interbib2007/Scripts/bibimport.py,v 1.12 2007/08/16 02:53:03 paepcke Exp $
# Description:  Importation of Bibtex Files to Python Objects
# Author:       Andreas Paepcke, Stanford University
# Created:      Sat Sep 23 09:12:10 1995
# Modified:     Wed Aug 15 19:21:25 2007 (Andreas Paepcke) paepcke@SNUGGLES
# Language:     Python
# Package:      N/A
# Status:       Experimental (Do Not Distribute)
#
# (c) Copyright 1995, Stanford University, all rights reserved.
#
################################################################################
#
# Revisions:
#
# Thu Aug 02 11:06:48 2007 (Andreas Paepcke) paepcke@EMIL
#  Added EndnoteImporter.
# Sun Jul 29 19:55:25 2007 (Andreas Paepcke) paepcke@EMIL
#  When a links entry is found in BibTex, we now automatically
#  add a 'url' field as well.
# Sun Jul 29 12:13:53 2007 (Andreas Paepcke) paepcke@EMIL
#  Added field entryType to each entry instance. This field holds the
#  bib type, like inproceedings, book, etc. for use by exporters.
# Sun Jul 22 07:55:18 2007 (Andreas Paepcke) paepcke@SNUGGLES
#  Converted to the new exception convention (no string exceptions; new calls for each)
# Jul 20 09:50:33 2007 (Andreas Paepcke) paepcke@DBPubs.Stanford.EDU
#  Migrated from deprecated regex to re.
# Wed May 13 13:09:04 1998 (Andreas Paepcke) paepcke@Haddock.Stanford.EDU
#  In parseValue(): Took care of fields that mix abbreviations and
#  strings, like: jul # "--" # aug.
# Mon Dec 30 09:25:20 1996 (Andreas Paepcke) paepcke@Haddock.Stanford.EDU
#  In deriveKey: when inventing a key, I used to create: bib-1, bib-2,
#  etc. This caused trouble when keys were char-set adjusted and the '-'
#  was xformed to a control char: I couldn't find bib-0 and would reassign
#  that key over and over.
# Mon Dec 30 09:08:17 1996 (Andreas Paepcke) paepcke@Haddock.Stanford.EDU
#  In refer data finding: if a date field had no digit at all, the
#  findYearInDate method returned 0 instead of -1. That was bad.
# Mon Nov 18 18:46:04 1996 (Andreas Paepcke) paepcke@Stingray
#  In parseDate: Changed to accept seconds since 1970
# Fri Oct 18 10:15:06 1996 (Andreas Paepcke) paepcke@Stingray
#  Added additional abbreviations.
# Wed Sep 18 16:30:58 1996 (Andreas Paepcke) paepcke@Walrus
#  Added support for Macintosh files.
# Wed Sep 11 08:10:47 1996 (Andreas Paepcke) paepcke@Walrus
#  Modified key generation to use 'bib-0', 'bib-1', etc. for keys because
#  the automatically generated ones are too ugly.
# Tue Sep 10 20:06:44 1996 (Andreas Paepcke) paepcke@Walrus
#  Made URLs in CItems be attached to biblio notes. Hot-linked the 'note' field
#  instead of the title field.
# Fri Sep  6 09:05:17 1996 (Andreas Paepcke) paepcke@Walrus
#  Prettyfied the key generator.
# Wed Aug 28 18:39:30 1996 (Andreas Paepcke) paepcke@Stingray
#  In dlraw importer: made publisher be booktitle when nothing else fit.
# Fri Aug 23 14:30:00 1996 (Andreas Paepcke) paepcke@Walrus
#  Moved import of any to init method of raw importer. This way it doesn't get
#  imported by the Web-based InterBib, reducing dependencies on the rest of the
#  testbed which is less stable.
# Mon Aug 12 14:44:17 1996 (Andreas Paepcke) paepcke@Walrus
#  In dlraw importer: threw out all the fields that are any.Undefined.
# Sat Aug 10 16:56:29 1996 (Andreas Paepcke) paepcke@Walrus
#  In dlraw importer, caused item properties that have read-only syntax (start/end
#  with '__' to be skipped. They used to cause errors.
# Fri Jul 12 08:42:06 1996 (Andreas Paepcke) paepcke@Walrus
#  Added permissions parm to interbib importer
# Thu Jul 11 19:02:26 1996 (Andreas Paepcke) paepcke@Walrus
#  Added lots of CS journal abbrevs.
# Wed Jul 10 09:44:12 1996 (Andreas Paepcke) paepcke@Walrus
#  Took out setEntries.
# Tue Jul  9 17:33:50 1996 (Andreas Paepcke) paepcke@Walrus
#  Added mechanism to declare duplicate key policy (as is in bibdb)
# Tue Jul  9 17:11:25 1996 (Andreas Paepcke) paepcke@Walrus
#  Added method setEntries
# Mon Jul  1 18:00:09 1996 (Andreas Paepcke) paepcke@Phoebe.Stanford.EDU
#  In raw DL importer: in process record: added check for field values being
#  None. Throwing out the ones I find.
# Mon Jul  1 17:56:54 1996 (Andreas Paepcke) paepcke@Phoebe.Stanford.EDU
#  In raw DL importer: turned URLs to 'links' entry *after* each record is
#  processed. This fixed problem with 'links' is non-translatable attr problem.
# Fri Jun 14 14:27:49 1996 (Andreas Paepcke) paepcke@Stingray
#  Adjusted raw dl converter to take lists of DLItems, rather than dicts of dicts.
# Tue Feb  6 22:46:06 1996 (Andreas Paepcke) paepcke@Stingray
#  Changed bibconvert's UnknownLatexControlSeq to
#  UnknownOrNonTranslateableLatexControlSeq
# Sun Jan  7 13:22:41 1996 (Andreas Paepcke) paepcke@AndreasHomePC
#  Replaced progress report 'Entries imported successfully...' with
#  'Entries parsed successfully...'
# Mon Jan  1 12:31:45 1996 (Andreas Paepcke) paepcke@AndreasHomePC
#  Made error and warning reports use bibutils.report()
# Mon Dec 25 21:26:39 1995 (Andreas Paepcke) paepcke@Stingray
#  Caught errors happening when trying to open a non-existing shelve
# Mon Dec 25 21:03:24 1995 (Andreas Paepcke) paepcke@Stingray
#  Made bibtex importer throw error when record key is missing.
# Mon Dec 18 16:29:22 1995 (Andreas Paepcke) paepcke@Phoebe.Stanford.EDU
#  Fixed error with 'conference' entry type.
# Tue Oct 24 18:41:10 1995 (Andreas Paepcke) paepcke@Phoebe.Stanford.EDU
#  Changed instantiation of AttributeConverter to RawDLDocAttributeConverter
################################################################################

# To document:
#   - Use of refer field '%1' for linking.
#   - Use '%F' for key info; that fits best with Endnote

# Public Classes and Methods:
#
#   BibtexImporter
#     attr: errors
#     attr: warnings
#     showErrors
#     showWarnings
#     importBibliography
#     reloadBibliography

# Limitations of BibTex entries:
# - Putting BibTex into an entry whose nativeCharSet is not BibTexCharSet
#   discards curly braces that force capitalization. This is no problem if
#   outputCharSet is not also BibTexCharSet. If it is, you'll have lost those
#   curlies. Fix: if you put in as Bibtex and take out as bibtex, set
#   nativeCharSet to BibTexCharSet. That will fix it, and it's faster
#   anyway.
# - Don't handle mathematical control sequences in BibTex entries.

# To do:
# - Improve presentation of BibEntry records in 'describe'
# - Since this was my first Python prog, I was still experimenting with
#   different styles of var names, etc. That should be cleaned to look
#   more like bibrefgen.

# Usage:
#
# bi = bibimport.BibtexImporter("/u/paepcke/tmp/hooplatest.bib")
# bi.importBibliography()   # Import the bib file (see verbose/quiet option below)
#                           # Duplicate key handling as per documented in bibdb.py
# bi.showErrors()           # List errors that occurred during import
# bi.showWarnings()         # List warnings that occurred during import
# en = bi['paep92a']
# en.describe()
# en.title
# en.author...
#   For reloading a bibl that was loaded before:
# bi.reloadBibliography()
#
#   Default for putting in entry field values is LaTex. By default this is
#   stored internally using the FrameMaker char set. When getting field
#   values, they are delivered in FrameMaker char set. You can change both
#   of these:
#   Specify a different output format for entry fields:
# entry.outputCharSet = LatexCharSet
#   Specify a different input format for entry fields:
# entry.inputCharSet = FrameCharSet
#   If you always put in as LaTex and pull out as LaTex, it is more efficient
#   (and less lossy of format specs such as \it) to set the native storage
#   mode to LaTex:
#   Specify a different format for entry fields to be stored internally
# entry.nativeCharSet = LatexCharSet
#
#   The importBibliography() method has options:
#      fileName = <file-name>: BibTex file name. Use if you didn't enter it
#                              as part of the instantiation as shown above.
#                              Use what was used in instantiation
#      verbose = [1|0]       : Set to 0 is you don't want to see errors/warnings
#                              runtime status printed to the screen. Default: 1
#   The reloadBibliography() method has options:
#      fileName = <file-name>
#      verbose =  [1|0]
#      eraseAbbrevs = [1|0]  : Clear out all BibTex string macro defs. Default: 0
#                              Duplicate string macros defs cause warnings which are
#                              harmless

import bibdb
import bibutils
import bibconvert
import USMarc
import string
import re
import time
import types
import shelve
import os
import sys
import tempfile

#####################################################
#
# Constants
#
#####################################################

# How authors are separated from one another internally in the author field:
authorSeparator = '|'

bibTexCommentChar = '%'
bibTexEscapeChar  = '\\'

# For the following regexp patterns, remember
# that the reserved metacharacters that need
# to be escaped are: . ^ $ * + ? { [ ] \ | ( )

# Pattern to match the BibTeX comment character '%'
# either at the start of a string, or in the middle,
# without escape. Matches:  "%foo bar" and "foo % bar",
# but not "foo \% bar" or "\% foo":

commentExtractPattern = """
           ([^%]*?[\{].*[}]|     # <any non-% chars>GREEDILY:{<any chars>}<any chars><not-\>% OR
           [^%]*?["].*["]|       # <any non-% chars>GREEDILY:"<any chars>"<any chars><not-\>% OR
           [^%]*?['].*['])       # <any non-% chars>GREEDILY:'<any chars>'<any chars><not-\>%
           .*?[^\\\\][%]|        # Any of the above,followed by <any chars><any non-\ char>%
	   [^"'\{]*?[^\\\\]?[%]  # Comment in a line that's NOT part of a BibTeX string:
	                         # <any non-"'{-chars> followed GREEDILY by either no char
	                         # or a char that not an escape backslash, followed by %
	   """

nonCommentExtractor = re.compile(commentExtractPattern, re.VERBOSE | re.DOTALL)
nonEscBibTeXComment = re.compile(r"""(?<=[^\\])[%]|^[%]""", re.VERBOSE)

# Helps determine whether a string is itself
# a string: "'foo'" ==> True
#           "{foo}" ==> True
#           '"foo"' ==> True
#           "foo"   ==> False:
strDetector = re.compile("""
			 \s*            # Skip leading whitespace
			 ([\{].*[}]|    # String delimited by {}?
			 ["].*["]|      # String delimited by ""?
			 ['].*['])      # String delimited by ''?
			 """, re.VERBOSE)

# The following four regex's use negative lookbehind assertions.
#   (?<=...)<re>: Match if <re> is NOT preceded by ...
# The "|^[<char>]" is needed each time, b/c the first 
# part of the pattern does not match if <char> is the
# first char on a line:

nonEscOpenBraceFinder   = re.compile(r"""(?<=[^\\])[{]|^[{]""") # A '{' NOT preceded by a backslash
nonEscCloseBraceFinder  = re.compile(r"""(?<=[^\\])[}]|^[}]""") # A '}' NOT preceded by a backslash
nonEscSingleQuoteFinder = re.compile(r"""(?<=[^\\])[']|^[']""") # A "'" NOT preceded by a backslash
nonEscDoubleQuoteFinder = re.compile(r"""(?<=[^\\])["]|^["]""") # A '"' NOT preceded by a backslash

# Regexps to find the LAST of '}', "'", or '"' in a string, given 
# that the string starts with '{', "'", or '"', respectively:

lastCloseBraceFinder    = re.compile(r"""^[{].*(?<=[^\\])[}]""")
lastSingleQuoteFinder   = re.compile(r"""^['].*(?<=[^\\])[']""")
lastDoubleQuoteFinder   = re.compile(r"""^["].*(?<=[^\\])["]""")

# Matches if a string's first non-white char is a BibTex comment char:
leadingCommentDetector  = re.compile(r"""^%|^\s*%""", re.VERBOSE)

# Matches if a comment is preceded only by a non-string.
# This is implemented as not allowing any of "{", "'", or '"'
# to precede the comment:

trailingCommentDetector = re.compile(r"""^[^{"'%]*%""", re.VERBOSE)

# Match a legal macro name at the start of a string. So,
# "foo, bar" matches up to the end of "foo". And ", blue"
# does not match. Nor does "   foo":

cleanMacroName = re.compile(r"""^[a-zA-Z0-9_]*""", re.VERBOSE)

OPEN_BRACE   = '{'
CLOSE_BRACE  = '}'
SINGLE_QUOTE = "'"
DOUBLE_QUOTE = '"'
NO_QUOTE     = ""


#####################################################
#
# Exceptions
#
#####################################################

class EntryParsingError(bibutils.InterbibException):
  pass

class DuplicateRecordKey(bibutils.InterbibException):
  pass

class RunnawayArgument(bibutils.InterbibException):
  pass

class IllegalRecordType(bibutils.InterbibException):
  pass

class IllegalFieldValue(bibutils.InterbibException):
  pass

#####################################################
#
# Class Definitions
#
#####################################################

# NOTE: this is an abstract class. 

class BibliographyImporter:

  importComplete = 0
  entries = {}
  warnings = []
  errors = []
  fd = 0
  abbrevs = {}


  # If fileName is non-empty, assume import will happen from a file.
  # If infoStr is non-empty, assume import will happen from a str.
  # If infoStruct is non-empty, the subclass will have to deal with
  # the structure that contains the info to import:

  def __init__(self, fileName='', infoStr='', infoStruct=None, verbosity=1):
    self.verbosity = verbosity
    self.entries = {}
    self.fileName = fileName
    self.infoStr  = infoStr
    self.infoStruct = infoStruct
    self.duplicateRecovery = bibdb.autoResolveRecordConflicts
    self.fd = 0
    self.tmpName = ''
    self.keyGenerator = bibutils.KeyGenerator()
    self.initAbbrevs()
    try:
      if fileName:
        self.fd = bibutils.StrOrFileFd(fileName=fileName)
	# We have to handle Macintosh files differently:
	self.checkForMacFile()
      elif infoStr:
	self.fd = bibutils.StrOrFileFd(str=infoStr)
    except IOError, info:
      if verbosity:
        bibutils.report("Object was created. But file not opened " +
		        str(info) +
		        ".\nUse the importBibliography's fileName parm to retry.")

  def initAbbrevs(self):
    self.abbrevs = {'jan':'January', 'feb':'February', 'mar':'March',
	            'apr':'April',   'may':'May',      'jun':'June',  'jul':'July',
	            'aug':'August',  'sep':'September','oct':'October',
	            'nov':'November','dec':'December',
		    'Jan':'January', 'Feb':'February', 'Mar':'March',
		    'Apr':'April',   'May':'May',      'Jun':'June',  'Jul':'July',
		    'Aug':'August',  'Sep':'September','Oct':'October',
    	            'Nov':'November','Dec':'December',
                    'acm':'Association of Computing Machinery',
		    'acmcs':'ACM Computing Surveys', 'acta': 'Acta Informatica',
		    'cacm': 'Communications of the ACM',
		    'cose': 'Proceedings of the International Conference on Software Engineering',
		    'dapda': 'Distributed and Parallel Databases',
		    'dbpd': 'Database Programming & Design',
		    'de': 'Proceedings of the Computer Science Internationl Conference on Database Engineering',
		    'dood': 'International Conference on Deductive and Object-Oriented Databases',
		    'ibmjrd': 'IBM Journal of Research and Development',
		    'ibmsj': 'IBM Systems Journal',
		    'ieeecomp': 'IEEE Computer',
		    'ieeese': 'IEEE Transactions on Software Engineering',
		    'ieeetc': 'IEEE Transactions on Computers',
		    'ieeetcad': 'IEEE Transactions on Computer-Aided Design of Integrated Circuits',
		    'ijcai': 'International Joint Conference on Artificial Intelligence',
		    'ijicis': 'International Journal of Intelligent & Cooperative Information Systems',
		    'ipl': 'Information Processing Letters', 'jacm': 'Journal of the ACM',
		    'jcss': 'Journal of Computer and System Sciences',
		    'oopsla': 'Proceedings of the Conference on Object-Oriented Programming Systems, Languages, and Applications',
		    'pods': 'Proceedings of the ACM SIGACT-SIGMOD Symposium on Principles of Database Systems',
		    'scp': 'Science of Computer Programming',
		    'sigmod': 'Proceedings of the ACM SIGMOD International Conference on Management of Data',
		    'sicomp': 'SIAM Journal on Computing',
		    'tcs': 'Theoretical Computer Science',
		    'tocs': 'ACM Transactions on Computer Systems',
		    'tods': 'ACM Transactions on Database Systems',
		    'tog': 'ACM Transactions on Graphics',
		    'tokde': 'IEEE Transactions on Knowledge Data Engineering',
		    'toms': 'ACM Transactions on Mathematical Software',
		    'toois': 'ACM Transactions on Office Information Systems',
		    'toplas': 'ACM Transactions on Programming Languages and Systems',
		    'tose': 'IEEE Transactions on Software Engineering',
		    'vldb': 'Proceedings of the International Conference on Very Large Databases'}

  def describe(self):
    bibutils.report('This object imports bibliographic citations into the '\
          'citation server.')

  # Make it so that importObject['paep88'] will return a bib entry:
  def __getitem__(self, key):
    return(self.entries[key])

  def showErrors(self):
    if self.errors:
      for error in self.errors:
        bibutils.report(error)
    else:
      bibutils.report("No errors.")

  def showWarnings(self):
    if self.warnings:
      for warn in self.warnings:
        bibutils.report(warn)
    else:
      bibutils.report("No warnings.")

  # Given the last-name-first name of the first author in self.firstAuthor,
  # construct a key:

  def deriveKey(self, entry):
    originalResolvePolicy = self.duplicateRecovery
    if self.firstAuthor:
      if len(self.firstAuthor) >= 4:
	keyRoot = self.firstAuthor[:4]
      else:
	keyRoot = self.firstAuthor
      if hasattr(entry, 'year'):
	key = keyRoot + entry.year[2:]
      else:
	key = keyRoot
    else:
      #************
      # Key generator or code here should try to use title for key gen!!
      key = 'bib'
      self.duplicateRecovery = bibdb.autoResolveRecordConflicts
      #key = self.keyGenerator.generateKey()

    if self.entries.has_key(key):
      if self.duplicateRecovery == bibdb.autoResolveRecordConflicts:
        for i in range(100000):
          newKey = key + str(i)
          if not self.entries.has_key(newKey):
            #self.warnings.append("Auto-resolved duplicate key '" +\
      	    #		   key +\
      	    #		   "' to new key '" +\
      	    #		   newKey +\
      	    #		   "'")
            key = newKey
            break
      else:
        self.errors.append("Key '" +\
      		      key +\
      		     "' is already in the bibliography database. Ignored")
        raise DuplicateRecordKey, key
    self.duplicateRecovery = originalResolvePolicy
    return(key)

  # Given an entry, look whether it has info in the special field
  # for links. Then set the entry's field to contain the proper
  # linking information:

  def processLinks(self, entry):

    # Process link information if any is there:
    linkInfo = bibutils.getEntryFldSafely(entry, bibconvert.linksFieldName)
    if linkInfo:
      linkObjs = []
      # Get array of strings, each with one link specification:
      try:
        allLinks = bibutils.extractLinks(linkInfo)
        for strLinkSpec in allLinks:

          # Get array: [<anchorFieldName>, <targetEnv>, <dest>]:
          linkSpecList = bibutils.parseLinkSpec(strLinkSpec)
          # Get a link object of the proper Class:
          linkObj = bibdb.makeLinkObject(linkSpecList[1])
          linkObj.anchorField = linkSpecList[0]
          linkObj.destination = linkSpecList[2]
          linkObjs.append(linkObj)

	  # If the link is a WWW link (as opposed to a Framemaker link),
	  # and if the entry doesn't already have a 'url'
	  # field, then add such a url field:

	  if (linkObj.system == 'www') and not \
	     bibutils.getEntryFldSafely(entry, bibconvert.urlFieldName):
	    entry[bibconvert.urlFieldName] = linkObj.destination

      except bibutils.BadLinkSpecFormat, info:
        self.warnings.append(info.message())
	# Delete the faulty link info from the links field so as not to
	# cause trouble later:
  	# This does not work for unknown reasonse: delattr(entry, bibconvert.linksFieldName)
	setattr(entry, bibconvert.linksFieldName, None)
      except RunnawayArgument, info:
        self.warnings.append(info.message())
      if linkObjs:
        setattr(entry, bibconvert.linksFieldName, linkObjs)

  # Import all records. If 'verbose' is true, output intermediate status every
  # few seconds, and print errors if they occur. Returns number of records
  # processed, including string macro definitions. Use methods showErrors() and
  # showWarnings() to get a list of errors and warnings afterwards.

  def importBibliography(self,
			 fileName = "",
			 infoStr = "",
			 infoStruct = None,
			 duplicateRecovery = bibdb.autoResolveRecordConflicts,
			 verbose = 1):
    self.duplicateRecovery = duplicateRecovery
    self.fileName = fileName
    self.infoStr  = infoStr
    self.infoStruct = infoStruct
    if not self.fd or \
       (fileName and fileName <> self.fd.name()) or \
       (infoStr and infoStr <> self.fd.str):
      try:
        if fileName:
          self.fd = bibutils.StrOrFileFd(fileName=fileName)
	  # We have to handle Macintosh files differently:
	  self.checkForMacFile()
        else:
	  self.fd = bibutils.StrOrFileFd(str=infoStr)
	self.importComplete = 0
      except IOError, info:
	if verbose:
          bibutils.report("Could not access bibliography file. " + str(info))
	return
    self.errors = []
    self.warnings = []
    warningsIndx = 0
    if verbose:
      startTime = time.time()
    recsProcessed = 0
    if self.importComplete and verbose:
      # This same file was imported before. Warn to use reload version:
      bibutils.report("This bibliography file was already imported. Use reloadBibliography to do it again")
    while not self.importComplete:
      try:
	self.importRecord(recsProcessed)
	recsProcessed = recsProcessed + 1
	if verbose and len(self.warnings) > warningsIndx:
	  for warning in self.warnings[warningsIndx:]:
	    bibutils.report("Warning: " + warning)
	  warningsIndx = len(self.warnings)
	if verbose and (time.time() - startTime) > 5.0:
	  bibutils.report('     Imported ' + str(recsProcessed) + ' records...')
	  startTime = time.time()
      except EntryParsingError:
	if verbose:
	  bibutils.report('*** Error: ' + self.errors[-1])
	continue
      except DuplicateRecordKey:
	if verbose:
	  bibutils.report('*** Error: ' + self.errors[-1])
	continue
      except RunnawayArgument:
	if verbose:
	  bibutils.report('*** Error: ' + self.errors[-1])
      except IllegalRecordType:
	if verbose:
	  bibutils.report('*** Error: ' + self.errors[-1])
      except IllegalFieldValue:
	if verbose:
	  bibutils.report('*** Error: ' + self.errors[-1])
      except bibconvert.UnsupportedCharSet, charSet:
	self.errors.append('Unsupported character set: ' + charSet.message())
	if verbose:
	  bibutils.report('*** Error: ' + self.errors[-1])
      except bibconvert.UnknownOrNonTranslateableLatexControlSeq, cntSeq:
	self.errors.append('Unknown or non-translateable LaTeX control sequence in entry '+\
			   self.currentRecKey + ':' + cntSeq.message())
	del(self.entries[self.currentRecKey])
	if verbose:
	  bibutils.report('*** Error: ' + self.errors[-1])
    if verbose:
      bibutils.report("Total import errors: " + str(len(self.errors)))
      bibutils.report("Total import warnings: " + str(len(self.warnings)))
      bibutils.report("Entries parsed successfully: " + str(len(self.entries) ))
    self.cleanup()
    return(len(self.entries))
    # return(recsProcessed)

  def reloadBibliography(self, verbose=1, eraseAbbrevs = 0):
    self.importComplete = 0
    if self.fd:
      self.fd.seek(0)
    self.entries = {}
    if eraseAbbrevs:
      self.initAbbrevs()
    return(self.importBibliography(fileName = self.fileName,
				   infoStr  = self.infoStr,
				   infoStruct = self.infoStruct,
				   verbose = verbose))

  def lineEmpty(self, ln):
    return(len(string.strip(ln)) == 0)

  # Mac files have lines separated by '\015' (^M) only, while Unix is '\012' (^J).
  # If the file we are getting is a Mac file, we work in memory:

  def checkForMacFile(self):
    fileSize = self.fd.len()
    # For Mac files, the following will be the entire thing:
    firstLn = self.fd.readline()
    if len(firstLn) == fileSize and\
       string.find(firstLn, '\015'):
      # Don't need the file anymore, because the whole contents is in the string:
      self.fd.close()
      #
      unixFormatStr = re.sub('\015', '\012', firstLn) 
      # Using the string as a fake file works fine, but is slower than doing a
      # temp file:
      #self.fd = bibutils.StrOrFileFd(str=unixFormatStr)
      # Initialize the temp file name generation:
      tempfile.template = 'interbibTmp-'
      self.tmpName = tempfile.mktemp()
      self.fd = open(self.tmpName, 'w')
      self.fd.write(unixFormatStr)
      self.fd.close()
      self.fd = self.fd = bibutils.StrOrFileFd(fileName=self.tmpName)
    else:
      self.fd.seek(0)

  def cleanup(self):
    if self.tmpName:
      if os.path.isfile(self.tmpName):
	os.system('rm ' + self.tmpName)

# End of class BibliographyImporter

class BibtexImporter(BibliographyImporter):

  rawAuthorSeparator = " and " # How authors are seped in a raw record to import
  maxNesting = 10              # How many nesting levels of curly braces to allow
                               # before assuming unbalanced braces
  recordStartChar = '@'
  stringCloseQuotes = {"{": "}",  '"': '"', "'":  "'"}

  waitForStringClose = False

  # The 'url' field is not an official field, but we include it anyway.
  # (BibTeX is older than the Web):

  officialFldNames = ['address', 'annote', 'author', 'booktitle',
		      'chapter', 'edition', 'editor', 'howpublished',
   	  	      'institution', 'journal', 'key', 'month',
		      'note', 'number', 'organization', 'pages',
		      'publisher', 'school', 'series', 'title',
		      'type', 'volume', 'year', 'url']
  currentRecKey = "" # So that error msgs from deep in recursion can be nice

  def __init__(self, fileName='', infoStr='', infoStruct=None, verbosity=1):
    BibliographyImporter.__init__(self,
				  fileName=fileName,
				  infoStr = infoStr,
				  infoStruct = infoStruct,
				  verbosity=verbosity)

  # Giving the definition of a Bibtex string macro, isolate the
  # macro name from its definition. Enter the abbreviation into the
  # abbrevs dictionary. The expected input is of the form:
  #       'cacm = "Communications of the ACM"}\012'

  def processStringMacroDef(self, macroDef):
    # Isolate the string macro name and its expansion:
    nameExpansion = string.split(macroDef, "=")
    shortcutKey = string.strip(nameExpansion[0])
    for shortcutExpansion in string.split(nameExpansion[1], '"'):
      shortcutExpansion = string.strip(shortcutExpansion)
      if len(shortcutExpansion) > 0: break
    if self.abbrevs.has_key(shortcutKey):
      self.warnings.append('Duplicate string def; new replaces the old or built-in def: "' +\
			   self.currentRecKey +\
			   '"')
    self.abbrevs[shortcutKey] = shortcutExpansion

  def importRecord(self, recsProcessed):

    self.curRecKey = ""
    self.findRecordStart()
    openCurlies = 0
    openQuotes = 0
    openParens = 0
    rec = []
    lnIndx = 0
    # Peek at the first line to see whether the BibTex record is delimited
    # by curlies as is the custom, or with parens, which is legal as well.
    # It just makes the loop faster:
    ln = self.fd.readline()
    if (string.find(ln, "(") > -1) and string.find(ln, "{") == -1:
      parenDelimited = 1
    else:
      parenDelimited = 0
    self.fd.seek(-len(ln), 1)
    ln = ""
    # Read one whole bibtex record into a struct:
    while 1:
      # Read one 'logical' line: multi-line fields are made into 
      # one string, and all BibTeX comments are stripped.
      # If we are in the middle of a multi-line parse, need to append
      # and insert a space to replace the newline:

      if lnIndx > 0:
        ln = ln + " " + self.readOneLogicalLine()
      else:
	ln = self.readOneLogicalLine()
      newCloseCurlies = bibutils.countMatches('}', ln[lnIndx:])\
		        - bibutils.countMatches('\}', ln[lnIndx:])
      openCurlies = openCurlies\
		    + bibutils.countMatches('{', ln[lnIndx:])\
		    - bibutils.countMatches('\{', ln[lnIndx:])\
		    - newCloseCurlies

      # If entry is paren-delimited, we need to track parentheses:
      if parenDelimited:
        newCloseParens = bibutils.countMatches(')', ln[lnIndx:])\
		          - bibutils.countMatches('\)', ln[lnIndx:])
        openParens = openParens\
		      + bibutils.countMatches('(', ln[lnIndx:])\
		      - bibutils.countMatches('\(', ln[lnIndx:])\
		      - newCloseParens

      if (openCurlies < 0) or (openParens < 0):
	self.errors.append(self.mkErrorMsg("Unexpected open brace or parenthesis.", 
					   rec, recsProcessed))
	raise RunnawayArgument, self.mkErrorMsg("Unexpected open brace/paren in:", 
						rec, recsProcessed)
	
      quotes = bibutils.countMatches('"', ln[lnIndx:]) -\
	       bibutils.countMatches('\\"', ln[lnIndx:])
      # If we were looking for a closing quote, then any odd number of
      # quotes will mean closure of the quotation. Otherwise, if there
      # is an odd number of quote marks, we are now in an open quotation:
      if openQuotes:
	openQuotes = openQuotes - quotes%2
      else:
	openQuotes = quotes%2
      if self.importComplete and openCurlies:
	self.errors.append(self.mkErrorMsg("End of file without closing brace.", 
					   rec, recsProcessed))
	raise RunnawayArgument, self.mkErrorMsg("Braces not closed in:", rec, recsProcessed)
      if self.importComplete and openQuotes:
	self.errors.append(self.mkErrorMsg("End of record/file without closing quote.", 
					   rec, recsProcessed))
	raise RunnawayArgument, self.mkErrorMsg("Quotation not closed in:", rec, recsProcessed)
      # Open parens at end of file are a problem only if the BibTex record
      # was delimited by parentheses instead of braces. Else ignore this condition
      if parenDelimited and self.importComplete and openParens:
	self.errors.append(self.mkErrorMsg("End of file without closing parenthesis.", 
					   rec, recsProcessed))
	raise RunnawayArgument, self.mkErrorMsg("Parentheses not closed in:", 
						rec, recsProcessed)
      if (openCurlies > 1) or openQuotes:
	if openCurlies > self.maxNesting:
	  self.errors.append(self.mkErrorMsg("End of file without closing brace.", 
					     rec, recsProcessed))
	  raise RunnawayArgument, self.mkErrorMsg("End of file without closing brace.", 
						  rec, recsProcessed)
	lnIndx = len(ln)
	continue
      rec.append(ln)
      ln = ""
      lnIndx = 0
      if ((not parenDelimited) and (openCurlies == 0)) or\
	 (parenDelimited and openParens == 0):
	# Found this record's closing curly brace or /paren. All done, 
	# except if there was a runnaway string:
	if openQuotes:
	  self.errors.append(self.mkErrorMsg("End of record/file without closing quote.", 
					     rec, recsProcessed))
	  raise RunnawayArgument, self.mkErrorMsg("Quotation not closed in:", rec, recsProcessed)
	break
    # The look-ahead may have overshot the closing curly/paren. But it would
    # not have eaten into the next record. Just into any comments after the
    # record. So we discard the tail end of the last record, including the
    # closing curly or paren:

    if parenDelimited:
      closingPosition = string.rfind(rec[-1], ")")
    else:
      closingPosition = string.rfind(rec[-1], "}")
    rec[-1] = rec[-1][:closingPosition]

    # Fix entries where the bibtex entry body starts at the same line as the
    # entry key, as in: @book{hass90,  author = "Scott Hassan"

    firstLn = rec[0]
    commaPos = string.find(firstLn, ',')
    if commaPos > -1:
      restOfLine = string.strip(firstLn[commaPos+1:])
      if restOfLine:
	rec = [firstLn[:commaPos]] + [restOfLine] + rec[1:]

    self.processRecord(rec)

  # Reads one whole field of one record: reads lines until it finds the end
  # of the file, a line with an equal sign (indicating a new field line), or
  # the start of a new record. Returns one line, with any overflow lines
  # appended to the end of the first physical line. This is a rough cut at
  # partitioning into fields:

  def readOneLogicalLine(self):
    ln = self.getNonEmptyLine()
    if self.importComplete:
      return("")
    else:
      return(string.strip(ln) + self.readNextLineTrial())

  def readNextLineTrial(self):
    ln = self.getNonEmptyLine()

    if self.importComplete:
      return("")
    if string.strip(ln)[0] == self.recordStartChar or "=" in ln:
  
      # backtrack: we've hit the last line of the entry
      self.fd.seek(-len(ln), 1)
      return("")
    else:
      # Strip spaces and put a space at the start of the line (because it
      # will be merged with the line before it):
      ln = " " + string.strip(ln)
      return(ln + self.readNextLineTrial())

  # Return next non-empty line from file.
  # If all of the line is a BibTeX comment, the
  # line is skipped:

  def getNonEmptyLine(self):

    while 1:

      ln = self.fd.readline()

      if len(ln) == 0:  # EOF
        self.importComplete = 1
        return("")

      # The following used also to skip lines starting
      # with a BibTeX comment char (%). Like this:
      #     or (ln.lstrip()[0] == bibTexCommentChar):...
      # But WRONG: you might have a multi-line string
      # in which one of the non-first lines starts with
      # a percent sign... You have to make the comment 
      # distinction later.

      # Skip newline-only:
      if (self.lineEmpty(ln)):
	continue

      return(ln)

  def mkErrorMsg(self, errMsg, rec, recsProcessed):
    
    # Construct and return a nice error msg for errors
    # during import of a record. Takes a string, an array
    # that contains the lines of the record that were 
    # read so far, and the number of records that have
    # been parsed successfully so far.

    if (len(rec) == 0):
      if self.currentRecKey:
	return(errMsg + " " + \
	       "Offending record unavailable. " + \
	       str(recsProcessed) + \
	       " records processed so far. \nPrevious record: " + \
	       self.currentRecKey + ".")
      else:
	return(errMsg + " " + \
	       "Offending record unavailable. " + \
	       str(recsProcessed) + \
	       " records processed so far.")
	
    else:
      if self.currentRecKey:
	return(errMsg + " " + \
	       "Offending record: " + \
	       rec[0] + ". " + \
	       str(recsProcessed) + \
	       " records processed so far. Previous record: " + \
	       self.currentRecKey + ".\n")
      else:
	return(errMsg + " " + \
	       "Offending record unavailable. " + \
	       str(recsProcessed) + \
	       " records processed so far.")


  def processRecord(self, prec):

    # isolate the type of record: @string vs. @book, etc.  from the first
    # line which is something like 'article{paep88\012' or '@string{cacm =
    # ...}'. the result of the split will be something like:
    # ['article', 'paep88\012'] or ['string', 'cacm = ...']. I some cases,
    # bibliographies may use regular parens instead of braces to delimit
    # entries:

    if string.find(prec[0], '{') > -1:
      recordId = string.split(prec[0], '{')
    elif string.find(prec[0], '(') > -1:
      recordId = string.split(prec[0], '(')
    else:
      self.errors.append("Cannot find record type/key in entry AFTER: '" + \
			 str(self.currentRecKey) + \
			 "'")
      raise EntryParsingError, "Parsing error"
    # get the record type (the leading '@' is already taken out)
    recordType = string.lower(recordId[0])

    if recordType == 'string':

      self.currentRecKey = 'String macro: ' + recordId[1]
      self.processStringMacroDef(recordId[1])
    elif recordType == 'preamble':
      return
    else:
      # Remember ID of record we are working on so error
      # msgs further down can be nice:
      self.currentRecKey = string.strip(recordId[1])
      if not self.currentRecKey:
        self.errors.append("Cannot find record key in: '" + prec[0])
	raise EntryParsingError
      # If the last char of the value is a comma, that's the field delimiter.
      # Strip it:
      if self.currentRecKey[-1] == ',':
        self.currentRecKey = self.currentRecKey[:-1]
      if self.entries.has_key(self.currentRecKey):
	self.currentRecKey = self.handleDuplicateEntry(self.currentRecKey)
	if not self.currentRecKey:
	  return
      # pass type of record, unique record key (w/o trailing blanks/nls), etc:
      self.processRegularEntry(recordType, self.currentRecKey, prec)

  # Depending on the duplicate detection policy, resolve or don't, and return
  # the key to use if everything is resolved, else, return 0:

  def handleDuplicateEntry(self, key):
    if self.duplicateRecovery == bibdb.overwriteDuplicates:
      self.warnings.append("Overwriting previous entry '" + key + "'")
      if self.verbosity:
        bibutils.report('Warning: ' + self.warnings[-1])
      return(key)
    else:
      if self.duplicateRecovery == bibdb.autoResolveRecordConflicts:
        for i in range(100000):
          newKey = key + '-' + str(i)
          if not self.entries.has_key(newKey):
    	    self.warnings.append("Auto-resolved duplicate key '" +\
    			         key +\
    			         "' to new key '" +\
    			         newKey +\
    			         "'")
    	    if self.verbosity:
    	      bibutils.report('Warning: ' + self.warnings[-1])
    	    return(newKey)
	return(0)
      elif self.duplicateRecovery == bibdb.discardDuplicates:
        self.warnings.append("Discarding duplicate entry '" + key + "'")
        if self.verbosity:
          bibutils.report('Warning: ' + self.warnings[-1])
        return(0)
      else:
        self.errors.append("Key '" +\
    			key +\
    			"' is already in the bibliography database. Ignored.")
        raise DuplicateRecordKey, key
    

  def processRegularEntry(self, prectype, pkey, prec):

    # Create an instance of the proper bib entry class:
    # Check whether this record type is legal (if not, this generates either
    # a TypeError or a NameError:

    try:
      # Construct the entry class from the module name and the type of record:
      entryClass = "bibdb." + prectype
      entry = eval(string.lower(entryClass))()
    except TypeError:
      self.errors.append("Type '" +\
			  prectype +\
			 "' is not a legal BibTex record. Entry key: "+\
			 pkey)
      raise IllegalRecordType, prectype
    except AttributeError:
      self.errors.append("Type '" +\
			  prectype +\
			 "' is not a legal BibTex record. Entry key: "+\
			 pkey)
      raise IllegalRecordType, prectype
    # Remember the key (ex paep98):
    entry[bibconvert.keyFieldName] = pkey
    # Remember the BibTeX record type (ex. 'inproceedings'):
    entry[bibconvert.typeFieldName] = prectype
    self.entries[pkey] = entry

    # Go through each line of the entry (except for the field-type/key):
    for entryLn in prec[1:]:
      entryLn = string.strip(entryLn)
      if not entryLn:
	continue
      # Take out leading or trailing field separator commas:
      if entryLn[0] == ',':
	entryLn = entryLn[1:]
      elif entryLn[-1] == ',':
	entryLn = entryLn[:-1]
      # Separate Bibtex field name from value:
      sepIndx = string.find(entryLn, "=")
      fldName = string.strip(entryLn[0:sepIndx])

      commentStart = self.findNonEscCommentChar(fldName)
      if commentStart > -1:
	continue

      fldValue = string.strip(entryLn[1 + sepIndx:])

      fldName = self.parseFldName(fldName)
      if fldName == 'author':
	fldValue = self.parseAuthors(fldValue)
      else:
        fldValue = self.parseValue(fldValue)

      # Convert field name 'doi' (document object identifier)
      # into 'url':

      if fldName == 'doi':
	fldName = bibconvert.urlFieldName

      # For fields like: url = "\url{http://foo/bar}",
      # only store the URL itself. It'll be added in
      # if we output to BibTeX:

      if fldName == bibconvert.urlFieldName:
	urlValue = bibutils.findEmbeddedURL(fldValue)
	if urlValue:
	  fldValue = urlValue
	else:
	  self.errors.append("Bad URL: '" + fldValue + "' in '" + pkey + "'")
	  raise IllegalFieldValue, "Bad URL: '" + fldValue + "'"

      setattr(entry, fldName, fldValue)
      if entry.warnings:
	# Add information about which field the warnings occurred in:
	for warn in entry.warnings:
	  self.warnings.append(warn + " in '" + pkey + "'")
    self.processLinks(entry)

  # Takes a raw fld name, like '   ,author  ' and returns a clean
  # substring of it:
  def parseFldName(self, rawFldName):
    fldName = string.lower(rawFldName)
    return(fldName)

  # Takes a list of 'and' separated authors and returns a sequence of the
  # authors:

  def parseAuthors(self, rawFldValue):

    # Strip the quotes:
    rawFldValue = rawFldValue[1:-1]
    # Break out the individual authors, getting a tuple of strings:
    rawFldValue = string.splitfields(rawFldValue, self.rawAuthorSeparator)
    # Make the author tuple into a separated string:
    return(string.joinfields(rawFldValue, authorSeparator))

  # Takes the value part of a bibtex field (i.e. everything after the '='
  # sign. If that's a non-quoted string, it expands it as a string macro.
  # Otherwise it strips the quote marks or curly braces around it.
  # Note, a little known BibTeX feature is that you can mix strings and
  # abbreviations if you put '#' around them. Like: month = jul # "--" # sep.
  # We handle that:

  def parseValue(self, rawEntireFldValue):
   
    if not rawEntireFldValue:
      return("")
   
    res = ""
    quoteType = None

    # Get list of strings or macro names. So a field value like
    #    jul # "--" # 'sep' # {blue\`{e}ll # {foo#bar}:
    # turns into: [jul, "--", 'sep', {blue\`{e}ll}, {foo, bar}]:

    chunksList = rawEntireFldValue.split('#')
    
    # In most cases field values will just be something
    # like {foo is bar}, in which case the chunks list
    # will be length 1. To speed things up, we treat this
    # case right here, never entering the loop below:
    
    if len(chunksList) == 1:
      if strDetector.match(rawEntireFldValue):
	# A string of some kind ({...}, "...", etc.; 
	# strip the quotes and return that:

	return self.purgeCommentsFromStr(rawEntireFldValue.strip())[1:-1]

      return self.resolveStrMacro(self.purgeCommentsFromStr(rawEntireFldValue))
      
    # More complicated: Have #-concatenated mix of strings
    # and macros. Look at each chunk in the list in turn.
    # As in comment above: strings might be ripped apart
    # by embedded '#' chars:

    for chunk in chunksList:

      if not chunk:
	continue

      # See whether chunk is a macro, or a str
      # delimited by "{}", '""', or "''". We determine
      # that from the chunk that contains the string's
      # start. Var quoteType being None signals that
      # we are starting a macro or string with this
      # chunk:

      if quoteType == None:
	# *Start* of a string or macro name.
	quoteType = chunk.strip()[0]

	if (not quoteType in ['{', "'", '"']):

	  # Check for comment and cut it out since we're 
	  #not within a string now:

	  res = res + self.resolveStrMacro(self.purgeCommentsFromStr(chunk))
	  quoteType = None
	  continue

	# Not a macro or unquoted number; initialize info 
	# that keeps track of when string is properly closed:
	
	if quoteType == OPEN_BRACE:
	  closeQuoteType = CLOSE_BRACE
	else:
	  closeQuoteType = quoteType

	# Remove opening spaces and string quote char:
	chunk = chunk.lstrip()[1:]

	openQuoteCnt = 1
	closeQuoteCnt = 0

      # Count *non-escaped* string openings an closings:

      openQuoteCnt  = openQuoteCnt  + self.cntNonEscapedQuotes(chunk, quoteType)
      closeQuoteCnt = closeQuoteCnt + self.cntNonEscapedQuotes(chunk, closeQuoteType)

      if ((quoteType == OPEN_BRACE) and (openQuoteCnt == closeQuoteCnt)) or\
	 ((quoteType != OPEN_BRACE) and (openQuoteCnt > 1)):
	# Found close of string that might have been ripped apart by an 
	# embedded '#' char:

	res = res + chunk.rstrip()[:-1];
	quoteType = None
	continue

      res = res + chunk + '#'
      continue

    return res

  # Given a non-quoted BibTeX string macro name, 
  # return the expansion, or raise an exception.
  # We strip spaces around the incoming string,
  # if there are any:

  def resolveStrMacro(self, theStr):

    theStr = theStr.strip()

    # Try to get the expansion:

    try:
      return self.abbrevs[theStr]
    except:
      pass

    # String is not a valid abbrev. Allow numbers, like
    # dates without quotes around them:

    if bibutils.digitsOnly(theStr):
      return theStr
    else:
      self.errors.append("Unknown abbreviation '" +\
			 theStr + \
			 "' in " \
			 + self.currentRecKey)
      del(self.entries[self.currentRecKey])
      raise EntryParsingError, "Unknown abbreviation '" + theStr + "'"
    


  def cntNonEscapedQuotes(self, theStr, quoteType=OPEN_BRACE):

    if quoteType == OPEN_BRACE:
      return len(nonEscOpenBraceFinder.findall(theStr))
    elif quoteType == CLOSE_BRACE:
      return len(nonEscCloseBraceFinder.findall(theStr))
    elif quoteType == SINGLE_QUOTE:
      return len(nonEscSingleQuoteFinder.findall(theStr))
    elif quoteType == DOUBLE_QUOTE:
      return len(nonEscDoubleQuoteFinder.findall(theStr))
    else:
      raise ValueError, "Illegal string quote type specifier: " + str(quoteType)

  def findNonEscCommentChar(self, theStr):
    matchObj = nonEscBibTeXComment.search(theStr)
    if matchObj:
      return matchObj.start()
    else:
      return -1

  def findRecordStart(self):
    escapeOn = False
    
    while 1:

      chr = self.fd.read(1)

      if chr == bibTexEscapeChar:
	escapeOn = True
	continue

      if (chr == bibTexCommentChar) and (not escapeOn):
	# Swallow rest of line as comment:
	self.skipComment()
	continue

      if (chr == BibtexImporter.recordStartChar) or\
	 (len(chr) == 0):
	break;

      escapeOn = False

  def skipComment(self):
    self.fd.readline()
    
  # Given a string, return that string with any
  # comment removed, or return the same string
  # if no comment is there.
  #
  # NOTE: this is not a smart function in that
  #       it won't detect nested strings that are
  #       left open, or embedded percent signs
  #
  #         {foo bl\`{e}me "bar"} % Comment
  #       will correctly remove '% Comment'.
  #       But:
  #         {foo bl\`{e}me "bar" % Comment
  #       will return {foo bl\`{e}me "bar"
  #       i.e. the comment will be cut:

  def purgeCommentsFromStr(self, theStr):

    if leadingCommentDetector.match(theStr):
      return ""

    theTightStr = theStr.lstrip()
    if theTightStr[0] == '{':
      theMatch = lastCloseBraceFinder.match(theTightStr)
    elif theTightStr[0] == '"':
      theMatch = lastDoubleQuoteFinder.match(theTightStr)
    elif theTightStr[0] == "'":
      theMatch = lastSingleQuoteFinder.match(theTightStr)
    else:
      trailingComment = trailingCommentDetector.match(theTightStr)
      if trailingComment:
	theTightStr = theTightStr[:trailingComment.end() - 1]

      # We may still have junk after the macro name,
      # like a comma that was protected from earlier
      # removal by a trailing comment, as in "month = jun,    % A comment"
	
      macroNameCapture = cleanMacroName.match(theTightStr)
      if macroNameCapture:
	return theTightStr[:macroNameCapture.end()]
      return theStr

    if not theMatch:
      return theStr

    return theTightStr[:theMatch.end()]

# end of class bibteximporter


# The following HeuristicClassification class is
# abstract. It provides to its child classes the
# ability to find a good bib entry class, given
# the fields. For instance: the Refer style doesn't
# tell you whether it's inproceedings, or a book,
# or a techreport (as the BibTeX style tells you).
# This class looks at the fields and tries to find
# a good match

class HeuristicClassification:

  def __init__(self):

    worstCase = None
    self.builtinAttributes = []
    self.citationClassHits = []
    self.urlAttributeName  = bibconvert.urlFieldName  # If doc has this attr, it will contain a URL
    self.bibEntryModule = 'bibdb'
    self.citationClasses = bibutils.allSubclassesOf(self.bibEntryModule, bibdb.BibEntry)
    if not self.citationClasses:
      raise ValueError, 'No citation classes to classify entries against.'

    # Make a list of counters and build the list of legal attributes:

    for className in self.citationClasses:

      fullClassName = self.bibEntryModule + '.' + className
      classObj = eval(fullClassName)

      for attr in (classObj.requiredFields + classObj.optionalFields):
	if attr not in self.builtinAttributes:
	  self.builtinAttributes.append(attr)

      self.citationClassHits.append(OverlapCounter(fullClassName))
      if className == 'misc':
	self.worstCase = self.citationClassHits[-1]

  def findBestBibEntryClass(self, fields):
    for overlapCounter in self.citationClassHits:
      overlapCounter.reset()
      
    for field in fields:
      for counter in self.citationClassHits:
	try:
	  translatedField = self.attrConverter.convertAttr(field)
	except bibconvert.NonTranslatableAttribute:
	  translatedField = field
	if translatedField in counter.requiredFields:
	  counter.hits = counter.hits + 1

    maximalFit = max(self.citationClassHits)
    if maximalFit.hits < 2:
      return(self.worstCase.className)
    else:
      if maximalFit.className == 'bibdb.conference':
	maximalFit.className = 'bibdb.inproceedings'
      return(maximalFit.className)


# This imports refer Files

class ReferImporter(BibliographyImporter, HeuristicClassification):

  # The 'FileName is really the dict of dicts:
  def __init__(self, fileName='', infoStr='', infoStruct=None, verbosity=1):

    BibliographyImporter.__init__(self,
				  fileName=fileName,
				  infoStr = infoStr,
				  infoStruct = infoStruct,
				  verbosity=verbosity)

    HeuristicClassification.__init__(self)
    self.attrConverter = bibconvert.ReferAttributeConverter()
    self.recordTypeConverter = bibconvert.RecordTypeNameConverterEndnoteInterbib()
    self.firstAuthor = ''

  def importRecord(self, recsProcessed):

    self.findRecordStart()
    if self.importComplete:
      return
    rec = []
    entryType = ''
    self.firstAuthor = ''

    explicitTypeInfoPresent = 0

    self.propFldsThisRecord = []

    # Read one whole refer record into a struct:
    while 1:
      ln = self.readOneLogicalLine()
      if not ln:
	# Found this record's end:
	break
      else:

	# Remember the fields we found. This allows us later to find the best
	# entry type by best-overlap-analysis, unless we can
	# determine the best type in this loop by looking at the field
	# entries. The start of the line is '%<oneLetterFldName>':

        try:
	  # Get the InterBib equivalent for %A, %T, etc:
  	  fldName = self.attrConverter.convertAttr(ln[1])

        except bibconvert.NonTranslatableAttribute:
	  fldName = ln[1]

	fldValue = string.strip(ln[2:])
        if fldName not in self.propFldsThisRecord:
  	  self.propFldsThisRecord.append(fldName)
	rec.append([fldName, fldValue])

        # Try to deduce a good entry type from the fields we discover:

	# Is there an explicit Refer field for the record type (for Endnote that's %0)?
	if (fldName == bibconvert.typeFieldName):
	  try:

	    # Try to get from the Refer (particularly Endnote) record type 
	    # (e.g. "Conference Proceedings") to the BibTeX equivalent 
	    # (e.g. "inproceedings"):

	    entryType = self.recordTypeConverter.getInterBibRecTypeFromEndnote(fldValue)

	    # Conveniently, we made subclasses of BibEntry that
	    # have the same names as BibTeX record types. So we
	    # can find the proper subclass to instantiate (later, in processRecord):
	    entryType = "bibdb." + entryType

	    explicitTypeInfoPresent = 1

	  except bibconvert.NonTranslatableAttribute:
	    pass
	  
	elif (not explicitTypeInfoPresent) and (ln[1] == 'R'):
	  entryType = self.bibEntryModule + '.' + 'techreport'
	elif (not explicitTypeInfoPresent) and (fldName == 'journal'):
	  entryType = self.bibEntryModule + '.' + 'article'

    self.processRecord(rec, entryType)
  
  # Given an array of pairs: [<translatedFldName>, <entry line>], create a
  # bib entry object of the best-fitting class, with a key that is derived
  # from the first author:
  
  def processRecord(self, rec, entryTypeName):

    # If the best entry type was discerned in importRecord above, use that
    # discovery. Else do some heuristics now that all field names are
    # available (not in this code below, but that's where it would
    # go). If that's not fruitful, compute the best entry type
    # by finding the maximum field name overlap:

    if not entryTypeName:
      entryTypeName = self.findBestBibEntryClass(self.propFldsThisRecord)

    # The entryTypeName is of the form "bibdb.inproceedings", which is
    # the name of the proper entry subclass to instantiate. Do that instantiation:

    entry = eval(entryTypeName)()

    # Store the entry's type (inproceedings, book, etc.) in the entry.
    # Just extract the subclass name, which is the entry kind (e.g. inproceedings):

    entry[bibconvert.typeFieldName] = entryTypeName.split(".")[1]

    # Go through each line of the entry:

    for nameEntryPair in rec:

      fldName  = nameEntryPair[0]
      fldValue = nameEntryPair[1]

      if fldName == 'author':
	self.handleAuthors(entry, fldValue)
      elif fldName == 'date':
	self.handleDate(entry, fldValue)

      if (fldName <> 'author') and (fldName <> 'date') and (fldName <> bibconvert.typeFieldName):
        setattr(entry, fldName, fldValue)

      if entry.warnings:
	# Add information about which field the warnings occurred in:
	for warn in entry.warnings:
	  self.warnings.append(warn + " in '" + fldName + ': ' + fldValue + "'")

    # Compute a good key:
    if (not hasattr(entry, bibconvert.keyFieldName)) or (not getattr(entry, bibconvert.keyFieldName)):
      setattr(entry, bibconvert.keyFieldName, self.deriveKey(entry))

    self.processLinks(entry)
    self.entries[getattr(entry, bibconvert.keyFieldName)] = entry

  def handleAuthors(self, entry, authorName):
    # Did we already have one author?:
    if self.firstAuthor:
      entry.author = entry.author + authorSeparator + authorName
    else:
      entry.author = authorName
      firstAuthorComponents = string.split(bibutils.getAuthorLastNameFirst(authorName),
					   ',')
      self.firstAuthor = string.lower(firstAuthorComponents[0])
      # Take out any apostrophes:
      self.firstAuthor = string.joinfields(string.splitfields(self.firstAuthor, "'"), '')

  def handleDate(self, entry, dateStr):
    # Is it a year spec only?
    if (len(dateStr) == 4) and bibutils.digitsOnly(dateStr):
      entry.year = dateStr
      return
    # Try to find the year:
    year = self.findYearInDate(dateStr)
    if year > -1:
      entry.year = year
	
  def findYearInDate(self, dateStr):

    # Look for a digit, repeated at least
    # and at most four times. However, this
    # will match: "abc20052006". The 'group'
    # method below, will pick out the first
    # four digits:

    theRE = re.compile("[0-9]{4,4}")
    theMatch = theRE.search(dateStr)
    if theMatch:
      return theMatch.group()
    else:
	return -1

  def findRecordStart(self):

    # Need to look for a '%' char in column 0:
    prevChr = self.fd.read(1)

    if len(prevChr) == 0:
      # End of file
      self.importComplete = 1
      return

    if (prevChr == '%'):
      # Beginning of file?
      if self.fd.tell() == 1:
	# Backtrack to before the opening '%':
	self.fd.seek(-1, 1)
	return
      else:
	# Go to the char before the '%':
	self.fd.seek(-2, 1)
	chr = self.fd.read(1)
	if chr == '\n':
	  return
	else:
	  # Go back past the opening '%':
	  self.fd.seek(1,1)

    while 1:
      chr = self.fd.read(1)

      if len(chr) == 0:
	self.importComplete = 1  # EOF
	break
      if (prevChr == '\n') and (chr == '%'):
	self.fd.seek(-1, 1)
	return
      else:
	prevChr = chr

  # Reads one whole field of one record: reads lines until it finds the end
  # of the file, or a line starting with '%', or an empty line. Returns
  # one line, with any overflow lines appended to the end of the first
  # physical line. This is a rough cut at partitioning into fields:

  def readOneLogicalLine(self):
    ln = self.getLine()
    if self.importComplete or not string.strip(ln):
      return("")
    else:
      return(string.strip(ln) + self.readNextLineTrial())

  def readNextLineTrial(self):
    ln = self.getLine()
    if self.importComplete:
      return("")
    strippedLn = string.strip(ln)
    if (not strippedLn) or strippedLn[0] == '%':
      # backtrack: we've hit the last line of the field
      self.fd.seek(-len(ln), 1)
      return("")
    else:
      # Strip spaces and put a space at the start of the line (because it
      # will be merged with the line before it):
      ln = " " + strippedLn
      return(ln + self.readNextLineTrial())
    
  def getLine(self):
    ln = self.fd.readline()
    if len(ln) == 0:  # EOF
      self.importComplete = 1
      return("")
    else:
      return(ln)

# end of class ReferImporter


class EndnoteImporter(ReferImporter):

  # The 'FileName is really the dict of dicts:
  def __init__(self, fileName='', infoStr='', infoStruct=None, verbosity=1):

    # Init parent classes:

    ReferImporter.__init__(self,
			   fileName=fileName,
			   infoStr = infoStr,
			   infoStruct = infoStruct,
			   verbosity=verbosity)

    self.attrConverter = bibconvert.EndnoteAttributeConverter()
    self.recordTypeConverter = bibconvert.RecordTypeNameConverterEndnoteInterbib()
    self.firstAuthor = ''

# end of class EndnoteImporter


# This imports a dictionnary of dictionaries with field/value pairs. It
# tries to deduce from the field names which kind of entry each inner dictionnary
# should be.

class HeuristicDLRawImporter(BibliographyImporter, HeuristicClassification):

  # The 'FileName is really the dict of dicts:
  def __init__(self, fileName='', infoStr='', infoStruct=None, verbosity=1):

    BibliographyImporter.__init__(self,
				  fileName=fileName,
				  infoStr = infoStr,
				  infoStruct = infoStruct,
				  verbosity=verbosity)

    HeuristicClassification.__init__(self)
    self.attrConverter = bibconvert.RawDLDocAttributeConverter()

  def importBibliography(self,
			 listOfDicts,
			 duplicateRecovery = bibdb.autoResolveRecordConflicts,
			 verbose = 1):

    self.errors = []
    self.warnings = []
    warningsIndx = 0

    if verbose:
      startTime = time.time()
    recsProcessed = 0

    if self.importComplete and verbose:
      # This same data structure was imported before. Warn to use reload version:
      bibutils.report("This bibliography file was already imported. Use reloadBibliography to do it again")

    for dict in listOfDicts:
      if self.importComplete:
        break
      try:
	self.processRecord(dict)
	recsProcessed = recsProcessed + 1
	if verbose and len(self.warnings) > warningsIndx:
	  for warning in self.warnings[warningsIndx:]:
	    bibutils.report("Warning: " + warning)
	  warningsIndx = len(self.warnings)
	if verbose and (time.time() - startTime) > 5.0:
	  bibutils.report('     Processed ' + str(recsProcessed) + ' records...')
	  startTime = time.time()

      except EntryParsingError:
	if verbose:
	  bibutils.report('*** Error: ' + self.errors[-1])
	continue
      except DuplicateRecordKey:
	if verbose:
	  bibutils.report('*** Error: ' + self.errors[-1])
	continue
      except RunnawayArgument:
	if verbose:
	  bibutils.report('*** Error: ' + self.errors[-1])
      except IllegalRecordType:
	if verbose:
	  bibutils.report('*** Error: ' + self.errors[-1])
      except bibconvert.UnsupportedCharSet, charSet:
	self.errors.append('Unsupported character set: ' + charSet)
	if verbose:
	  bibutils.report('*** Error: ' + self.errors[-1])
      except bibconvert.UnknownOrNonTranslateableLatexControlSeq, cntSeq:
	self.errors.append('Unknown or non-translateable LaTeX control sequence in entry '+\
			   self.currentRecKey + ':' + cntSeq.message())
        if self.entries.has_key(self.currentRecKey):
 	  del(self.entries[self.currentRecKey])
	if verbose:
	  bibutils.report('*** Error: ' + self.errors[-1])
    if verbose:
      bibutils.report("Total errors: " + str(len(self.errors)))
      bibutils.report("Total warnings: " + str(len(self.warnings)))
      bibutils.report("Entries parsed successfully: " + str(len(self.entries)))
    return(len(self.entries))
    # return(recsProcessed)

  def reloadBibliography(self, listOfDicts, verbose=1, eraseAbbrevs = 0):
    self.importComplete = 0
    self.entries = {}
    if eraseAbbrevs:
      self.initAbbrevs()
    return(self.importBibliography(listOfDicts, verbose))

  def processRecord(self, dict):
    
    import any
    dictKeys = []
    values = []
    self.firstAuthor = ''
    for key in dict.keys():
      value = dict.get_property_value_no_cnv(key)
      if value and value <> any.Undefined:
	dictKeys.append(key)
	values.append(value)

    try:
      publicationPos = dictKeys.index(USMarc.kPublication)
      pubValue = dict[USMarc.kPublication]
      if USMarc.kNumber in dictKeys or\
	 (USMarc.kVolume in dictKeys and\
	  USMarc.kNumber in dictKeys):
	dict['journal'] = pubValue
	dictKeys[publicationPos] = 'journal'
	values[publicationPos] = pubValue
      elif USMarc.kEdition in dictKeys:
	dict['publisher'] = dict[USMarc.kPublication]
	dictKeys[publicationPos] = 'publisher'
	values[publicationPos] = pubValue
      else:
	dict['booktitle'] = dict[USMarc.kPublication]
	dictKeys[publicationPos] = 'booktitle'
	values[publicationPos] = pubValue
    except ValueError:
      pass
    try:
      if not 'booktitle' in dictKeys and not 'journal' in dictKeys:
        abbrevPubPos = dictKeys.index(USMarc.kAbbreviatedPublication)
	dict['booktitle'] = dict[USMarc.kAbbreviatedPublication]
	dictKeys[abbrevPubPos] = 'booktitle'
	values[abbrevPubPos] = dict[USMarc.kAbbreviatedPublication]
    except ValueError:
      pass
    # If document has a URL:
    #  - If the entry as a 'note' field, add the URL to it. Else create a
    #    new 'note' field with the URL
    #  - Link the note field to the URL:
    try:
      urlPos = dictKeys.index(USMarc.kURL)
      if USMarc.kBibliographicNote in dictKeys:
	note = bibutils.joinSentences(dict[USMarc.kBibliographicNote], values[urlPos])
      else:
	note = values[urlPos]
      dictKeys.append('note')
      values.append(note)
      dictKeys.append(bibconvert.linksFieldName)
      values.append('note:www:' + values[urlPos])
    except ValueError:
      pass

    bestEntryClass = self.findBestBibEntryClass(dictKeys)

    # The bestEntryClass is of the form "bibdb.inproceedings", which is
    # the name of the proper entry subclass to instantiate. Do that instantiation:

    entry = eval(bestEntryClass)()

    # Store the entry's type (inproceedings, book, etc.) in the entry.
    # Just extract the subclass name, which is the entry kind (e.g. inproceedings):

    entry[bibconvert.typeFieldName] = bestEntryClass.split(".")[1]

    entry.inputCharSet = bibconvert.Latin1CharSet
    # To make error messages more clear:
    self.currentRecKey = 'New Entry'
    
    for i in range(len(values)):
      name = dictKeys[i]
      value = values[i]
      adjustedName = self.processAttrName(name)
      if adjustedName == 'date':
	month = self.parseDate(value)
	if month:
	  setattr(entry, 'month', month)
      elif adjustedName == 'author':
	setattr(entry, 'author', self.parseAuthor(value))
      else:
        adjustedValue = self.processAttrValue(adjustedName, value)
	try:
          setattr(entry, adjustedName, adjustedValue)
	# If an item property has a name that starts and ends with '__', it's read-only,
	# so just skip it:
	except TypeError:
	  continue
    key = self.deriveKey(entry)
    entry.key = key
    self.currentRecKey = key
    self.processLinks(entry)
    self.entries[key] = entry

  # If attr name is one of the built-in ones, return it. Else try to convert it
  # to a built-in one. If that works, return the translation. Else give up and
  # return the name unchanged. This case is separated out here to allow a stricter
  # control of attributes later on:

  def processAttrName(self, name):
    if name in self.builtinAttributes:
      return(name)
    else:
      try:
        adjustedName = self.attrConverter.convertAttr(name)
        return(adjustedName)
      except bibconvert.NonTranslatableAttribute:
	return(name)

  def processAttrValue(self, name, value):
    return(str(value))
  
  # For now: this works with date strs of the form: 'April 15'. Year is in another
  # field. We look for the month and pull it out as a full-length month name:

  def parseDate(self, date):
    # If date is an int, it's seconds since 1970. Just pull out the month:
    if type(date) == types.IntType:
      date = string.split(time.ctime(date), ' ')[1]
    else:
      # Make sure it's really a string!
      date = str(date)
    lowerCaseDate = string.lower(date)
    for monthPossibility in ['jan', 'feb', 'mar', 'apr', 'may', 'jun',
			     'jul', 'aug', 'sep', 'oct', 'nov', 'dec']:
      if string.find(lowerCaseDate, monthPossibility) > -1:
	return(self.abbrevs[monthPossibility])
    return('')

  # We expect author to be a list of strings:
  
  def parseAuthor(self, authorList):
    res = ''
    if not authorList:
      return(res)
    firstAuthorComponents = string.split(bibutils.getAuthorLastNameFirst(authorList[0]),
					 ',')
    self.firstAuthor = string.lower(firstAuthorComponents[0])
    # Take out any apostrophes:
    self.firstAuthor = string.joinfields(string.splitfields(self.firstAuthor, "'"), '')

    for author in authorList:
      if res:
        res = res + authorSeparator + author
      else:
	res = author
    return(res)

# end of class HeuristicKeyValueImporter


# This imports an existing shelve (i.e. a persistent db). It's a trivial
# thing. We just open the shelve and assign it to self.entries. This
# importer exists for symmetry with the other importers when databases
# are to be merged in bibdb. HOWEVER: Note that importBibliography does
# NOT return the number of records, because len(<a shelve>) can be very
# slow!

class InterbibImporter(BibliographyImporter):

  def __init__(self, fileName='', infoStr='', infoStruct=None, verbosity=1, permissions='w'):

    BibliographyImporter.__init__(self,
				  fileName=fileName,
				  infoStr = infoStr,
				  infoStruct = infoStruct,
				  verbosity = verbosity)
    self.permissions = permissions

  # We do NOT return the number of records, because len(<a shelve>) can be very
  # slow!

  def importBibliography(self,
			 duplicateRecovery = bibdb.autoResolveRecordConflicts,
			 verbose = 1):
    self.errors = []
    self.warnings = []
    if not os.path.isfile(self.fileName):
      self.errors.append("Could not open InterBib database '" + \
			 self.fileName +\
			 "'")
      if verbose:
        bibutils.report('*** Error: ' + self.errors[-1])
      raise IOError, self.errors[-1]
      
    # For existing shelves, we can open w/ r or w. But when we
    # need to create a shelve, that doesn't work:
    if os.path.isfile(self.fileName):
      permissions = self.permissions
    else:
      permissions = 'c'
    try:
      self.entries = shelve.open(self.fileName, permissions)
    except:
      try:
        self.errors.append("Could not open InterBib database '" + \
		     	   self.fileName +\
			   "': " +\
			   sys.exc_value)
      except:
        self.errors.append("Could not open InterBib database '" + \
		           self.fileName +\
			   "'")
      if verbose:
        bibutils.report('*** Error: ' + self.errors[-1])
      raise IOError, self.errors[-1]

class OverlapCounter:
  requiredFields = None
  hits  = 0
  className = ''

  def __init__(self, className):
    self.className = className
    classObj = eval(className)
    self.requiredFields = classObj.requiredFields
    self.hits = 0

  def reset(self):
    self.hits = 0
  
  def __cmp__(self, other):
    if self.hits < other.hits:
      return(-1)
    elif self.hits == other.hits:
      return(0)
    else:
      return(1)

# end of class OverlapCounter

