#! /usr/bin/python
################################################################################
#
# File:         nph-bibsearchcgi.py
# RCS:          $Header: /u/testbed/CVSROOT/dldev/src/interbib2007/Scripts/nph-bibsearchcgi.py,v 1.4 2007/07/27 16:30:44 paepcke Exp $
# Description:  HTTP InterBib search server
# Author:       Andreas Paepcke, Stanford University
# Created:      Thu Jul  4 12:55:49 1996
# Modified:     Fri Jul 27 09:21:28 2007 (Andreas Paepcke) paepcke@EMIL
# Language:     Python
# Package:      N/A
# Status:       Experimental (Do Not Distribute)
#
# (c) Copyright 1996, Stanford University, all rights reserved.
#
################################################################################
#
# Revisions:
#
# Sun Jul  7 10:35:54 1996 (Andreas Paepcke) paepcke@Stingray
#  Turned off verbosity in reference generator.
# Thu Jul  4 12:56:25 1996 (Andreas Paepcke) paepcke@Phoebe.Stanford.EDU
#  Where directory and file name are merged: Took care that no double '/'
#  was put between directory and file name, no matter what is passed as a
#  directory name
################################################################################

#
# nph-bibsearchcgi.py
#
# Arturo Crespo  (crespo@cs.stanford.edu)
#
# Description:
#   This module receives a set of cgi variables (which can be produced by
#   nph-bibsearchcgi0.py), perform the requested search, and returns the
#   result in a nice format to a HTML browser.
#   Why this filename starts with nph?  the reason is that the http server
#     display the information as soon as it becomes availabe for cgi
#     scripts starting with this prefix.
#

# constants
# this set of constants define the boundaries for multipart documents
csBoundary = '|-------------------InterBib Service------|'
csStartBoundary = '\n--' + csBoundary + '\n'
csStopBoundary = '\n--' + csBoundary + '--\n'
ciVerbosity = 0

import sys

# prepare to respond the request
# we do this here to be able to echo back errors when importing
# all needed modules
sys.stdout.write('HTTP/1.0 200 0K\n')
sys.stdout.write('Server: NCSA/1.4.1\n')
sys.stdout.write('Content-type: multipart/x-mixed-replace;boundary=' + csBoundary + '\n\n')
sys.stdout.write(csStartBoundary)
sys.stdout.write('Content-type: text/html\n\n')
sys.stderr = sys.stdout

import bibsearch
import cgi
import bibrefgen
import bibcompile
import os


#
# printHeader
#   IN:
#   OUT:
#   DESC: prints a header for the bibsearch form
#
def printHeader():
  sys.stdout.write("""
  <table border=0>
  <tr>
    <td valign=bottom> 
      <img src="gifs/interbibSearch.gif"
                      align=absmiddle
                      height=140 width=100
                      hspace=20
                      border=0/>

      <h1> 
        InterBib: Search Engine
      </h1>
    </td>
    <td>
      <img src="gifs/dllogo-smaller-xparent.gif" align=absmiddle/>
    </td>
  </tr>
  </table>
  <hr/>""")


#
# SearchAndPrintResults
#   IN:  filename  -  cgi variable for the bibliography to be searched
#        directory -  cgi variable for the directory of the bibliography
#        title     -  optional cgi variable for a title search string
#        author    -  optional cgi variable for a author search string
#   OUT:
#         

def SearchAndPrintResults():

  # read the cgi variables, if not present, enter debug mode and
  # assume some reasonable defaults
  try:
    cgiVars = cgi.parse()
  except:
    print "**** Debug Mode"
    cgiVars = {}
    cgiVars['filename'] = ["bib95"]
    cgiVars['directory'] = ["/u/testbed/public_html/interbib/bibliographies"]
    cgiVars['title']=["inter"]

  sys.stdout.write('<html><body bgcolor="#FFECD9">');
  printHeader()
  sys.stdout.write('<p></p>Accessing the Database(s)...\n<hr/>')
  sys.stdout.write('\n')
  sys.stdout.flush()

  if not cgiVars.has_key('filename'):
    print "<hr/>"
    print "<strong>ERROR: You must specify a database</strong>"
    print "<hr/>"
    sys.stdout.flush()
    sys.exit()

  if cgiVars.has_key('exactMatch') and cgiVars['exactMatch'][0] == 'on':
    exactMatch = 1
  else:
    exactMatch = 0
  if cgiVars.has_key('annote') and cgiVars['annote'][0] == 'on':
    annote = 1
  else:
    annote = 0
  if cgiVars.has_key('caseSensitive') and cgiVars['caseSensitive'][0] == 'on':
    caseSensitive = 1
  else:
    caseSensitive = 0
  if cgiVars.has_key('useIndex') and cgiVars['useIndex'][0] == 'on':
    useIndex = 1
  else:
    useIndex = 0
  if cgiVars.has_key('sortEntries') and cgiVars['sortEntries'][0] == 'on':
    sortEntries = 1
  else:
    sortEntries = 0
    
  # Get a reference generator object which takes entries and produces
  # platform-independent references:
  refGenerator = bibrefgen.ReferenceGenerator(verbosity = 0)

  # Get an object that compiles platform-independent references to HTML:
  HTMLCompiler = bibcompile.BibHTMLCompiler()

  sys.stdout.write('</html></body>');
  sys.stdout.write(csStartBoundary)
  sys.stdout.write('<html><body bgcolor="#FFECD9">');
  sys.stdout.write('\n')
  printHeader()
  sys.stdout.write('<h1>Results: </h1>')

  for i in range(0,len(cgiVars['filename'])):
    fullFile = os.path.join(cgiVars['directory'][0], cgiVars['filename'][i])
    cgiVars['filename'][i] = fullFile
    bs=bibsearch.bibSearch(verbosity=ciVerbosity)
    try:
      bs.openIndex( cgiVars['filename'][i], useIdx=useIndex, permissions='r')
    except bibsearch.IndexOld:
      sys.stdout.write("Index for database '" +\
		       cgiVars['filename'][i] +\
		       "' is old. Please notify InterBib administrator.")
      sys.stdout.flush()
      continue
    except bibsearch.IndexFileNotFound, info:
      try:
        bs.createIndex( cgiVars['filename'][i] )
        bs.openIndex( cgiVars['filename'][i], useIdx=useIndex, permissions='r')
      except:
        sys.stdout.write("Could not create index for database '" +\
		         str(cgiVars['filename'][i]) +\
		         "'. Ignoring this database.")
        sys.stdout.flush()
	continue

    if (cgiVars.has_key('Alldb')):
      result = bs.bibliography
    else:
      result = {}
      if (cgiVars.has_key('key')):
        result = bs.search( 'key' ,
			    cgiVars['key'][0],
			    exactMatch=exactMatch,
			    caseSensitive=caseSensitive,
			    useIndex=useIndex)
        bs.setBibliography(result)

      if (cgiVars.has_key('title')):
        result = bs.search( 'title',
			     cgiVars['title'][0],
			     exactMatch=exactMatch,
			     caseSensitive=caseSensitive,
			     useIndex=useIndex)
        bs.setBibliography(result)

 
      if (cgiVars.has_key('author')):
        result = bs.search( 'author',
			    cgiVars['author'][0],
			    exactMatch=exactMatch,
			    caseSensitive=caseSensitive,
			    useIndex=useIndex)
        bs.setBibliography(result)

      if (cgiVars.has_key('year')):
        result = bs.search( 'year' ,
			    cgiVars['year'][0],
			    exactMatch=exactMatch,
			    caseSensitive=caseSensitive,
			    useIndex=useIndex)
        bs.setBibliography(result)

      if (cgiVars.has_key('otherField')) and\
	 (cgiVars.has_key('otherFieldValue')):
        result = bs.search( cgiVars['otherField'][0] ,
			    cgiVars['otherFieldValue'][0],
			    exactMatch=exactMatch,
			    caseSensitive=caseSensitive,
			    useIndex=useIndex)
        bs.setBibliography(result)

    resultLength = len(result)
    if resultLength > 0:
      print "<strong>Results from database " +\
	    os.path.basename(cgiVars['filename'][i]) +\
	    ":</strong>"
    print "<ul>"

    resultKeys = result.keys()

    if sortEntries:
      import bibutils
      bibutils.sortDictBibEntriesByLastName(result, resultKeys)

    
    for key in resultKeys:
      reference = refGenerator.generateRef(result[key],
					   'GenericStyle',
					   annote=annote)
      print "<li>" + "<strong>[" + key + "]</strong> " +   \
  	    HTMLCompiler.compileRef(reference)[3:] + "</li><p></p>"

    print "</ul>"

    if resultLength == 0:
      print "<strong>No matches in database " +\
	    os.path.basename(cgiVars['filename'][i]) +\
	    ".</strong><p></p>"

  print "<hr/>"
  
#
#
# main
#
#

SearchAndPrintResults()
sys.stdout.flush()

